import typing as t

import marshmallow
import pydantic
import pytest
import sqlalchemy
import typesystem


@pytest.fixture(scope="function")
def puppy_schema(app):
    from flama import schemas

    if schemas.lib == pydantic:
        schema_ = pydantic.create_model(
            "Puppy", custom_id=(t.Optional[int], None), name=(str, ...), age=(int, ...), owner=(t.Optional[str], None)
        )
    elif schemas.lib == typesystem:
        schema_ = typesystem.Schema(
            fields={
                "custom_id": typesystem.Integer(allow_null=True),
                "name": typesystem.String(),
                "age": typesystem.Number(),
                "owner": typesystem.String(allow_null=True),
            }
        )
    elif schemas.lib == marshmallow:
        schema_ = type(
            "Puppy",
            (marshmallow.Schema,),
            {
                "custom_id": marshmallow.fields.Integer(allow_none=True),
                "name": marshmallow.fields.String(),
                "age": marshmallow.fields.Number(),
                "owner": marshmallow.fields.String(allow_none=True),
            },
        )
    else:
        raise ValueError("Wrong schema lib")

    return schema_


@pytest.fixture(scope="function")
async def puppy_model(app, client):
    table = sqlalchemy.Table(
        "puppy",
        app.sqlalchemy.metadata,
        sqlalchemy.Column("custom_id", sqlalchemy.Integer, primary_key=True, autoincrement=True),
        sqlalchemy.Column("name", sqlalchemy.String, nullable=False),
        sqlalchemy.Column("age", sqlalchemy.Integer, nullable=False),
        sqlalchemy.Column("owner", sqlalchemy.String, nullable=True, default=None),
    )

    async with app.sqlalchemy.engine.begin() as connection:
        await connection.run_sync(app.sqlalchemy.metadata.create_all, tables=[table])

    yield table

    async with app.sqlalchemy.engine.begin() as connection:
        await connection.run_sync(app.sqlalchemy.metadata.drop_all, tables=[table])



import asyncio
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import Flama, exceptions, types
from flama.lifespan import Lifespan


class TestCaseLifespan:
    @pytest.fixture
    def app(self):
        return MagicMock(Flama)

    @pytest.fixture
    def lifespan(self):
        return Lifespan(MagicMock())

    @pytest.mark.parametrize(
        ["states", "startup_side_effect", "shutdown_side_effect", "exception"],
        (
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.complete"}),
                        ],
                        types.AppStatus.READY,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.shutdown"}),
                        ],
                        [
                            types.Message({"type": "lifespan.shutdown.complete"}),
                        ],
                        types.AppStatus.SHUT_DOWN,
                    ),
                ),
                None,
                None,
                None,
                id="ok",
            ),
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.complete"}),
                        ],
                        types.AppStatus.READY,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.shutdown"}),
                        ],
                        [
                            types.Message({"type": "lifespan.shutdown.complete"}),
                        ],
                        types.AppStatus.SHUT_DOWN,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.complete"}),
                        ],
                        types.AppStatus.READY,
                    ),
                ),
                None,
                None,
                None,
                id="restart",
            ),
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.failed", "message": "Foo"}),
                        ],
                        types.AppStatus.FAILED,
                    ),
                ),
                Exception("Foo"),
                None,
                exceptions.ApplicationError("Lifespan startup failed"),
                id="startup_fail",
            ),
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.complete"}),
                        ],
                        types.AppStatus.READY,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.shutdown"}),
                        ],
                        [
                            types.Message({"type": "lifespan.shutdown.failed", "message": "Foo"}),
                        ],
                        types.AppStatus.FAILED,
                    ),
                ),
                None,
                Exception("Foo"),
                exceptions.ApplicationError("Lifespan shutdown failed"),
                id="shutdown_fail",
            ),
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.complete"}),
                        ],
                        types.AppStatus.READY,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message(
                                {
                                    "type": "lifespan.startup.failed",
                                    "message": "Trying to start application from 'AppStatus.READY' state",
                                }
                            ),
                        ],
                        types.AppStatus.READY,
                    ),
                ),
                None,
                None,
                exceptions.ApplicationError("Trying to start application from 'AppStatus.READY' state"),
                id="start_from_ready",
            ),
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.shutdown"}),
                        ],
                        [
                            types.Message(
                                {
                                    "type": "lifespan.shutdown.failed",
                                    "message": "Trying to shutdown application from 'AppStatus.NOT_STARTED' state",
                                }
                            ),
                        ],
                        types.AppStatus.READY,
                    ),
                ),
                None,
                None,
                exceptions.ApplicationError("Trying to shutdown application from 'AppStatus.NOT_STARTED' state"),
                id="shutdown_from_not_started",
            ),
            pytest.param(
                (
                    (
                        [],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.startup"}),
                        ],
                        [
                            types.Message({"type": "lifespan.startup.complete"}),
                        ],
                        types.AppStatus.READY,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.shutdown"}),
                        ],
                        [
                            types.Message({"type": "lifespan.shutdown.complete"}),
                        ],
                        types.AppStatus.SHUT_DOWN,
                    ),
                    (
                        [
                            types.Message({"type": "lifespan.shutdown"}),
                        ],
                        [
                            types.Message(
                                {
                                    "type": "lifespan.shutdown.failed",
                                    "message": "Trying to shutdown application from 'AppStatus.SHUT_DOWN' state",
                                }
                            ),
                        ],
                        types.AppStatus.SHUT_DOWN,
                    ),
                ),
                None,
                None,
                exceptions.ApplicationError("Trying to shutdown application from 'AppStatus.SHUT_DOWN' state"),
                id="shutdown_from_shutdown",
            ),
            pytest.param(
                (
                    (
                        [
                            types.Message({"type": "lifespan.unknown"}),
                        ],
                        [],
                        types.AppStatus.NOT_STARTED,
                    ),
                ),
                None,
                None,
                None,
                id="unknown_message",
            ),
        ),
        indirect=["exception"],
    )
    async def test_call(self, states, startup_side_effect, shutdown_side_effect, exception):
        receive_queue = asyncio.Queue()
        send_queue = asyncio.Queue()

        async def receive() -> types.Message:
            return await receive_queue.get()

        async def send(message: types.Message) -> None:
            await send_queue.put(message)

        app = Flama(docs=None, schema=None)
        sub_app = Flama(docs=None, schema=None)
        app.mount("/foo", sub_app)

        lifespan = app.router.lifespan

        with (
            exception,
            patch.object(lifespan, "_startup", side_effect=startup_side_effect),
            patch.object(lifespan, "_shutdown", side_effect=shutdown_side_effect),
        ):
            for receive_messages, send_messages, app_status in states:
                for m in receive_messages:
                    await receive_queue.put(m)
                    await lifespan(types.Scope({"app": app, "type": "lifespan"}), receive, send)

                assert [await send_queue.get() for _ in range(len(send_messages))] == send_messages
                assert app._status == app_status
                assert sub_app._status == app_status

    @pytest.mark.parametrize(
        ["child_lifespan"],
        (
            pytest.param(MagicMock(), id="lifespan"),
            pytest.param(None, id="no_lifespan"),
        ),
    )
    async def test_startup(self, app, lifespan, child_lifespan):
        lifespan.lifespan = child_lifespan
        foo = AsyncMock()
        app.events = MagicMock()
        app.events.startup = [foo]

        await lifespan._startup(app)

        assert foo.await_args_list == [call()]
        if child_lifespan:
            assert lifespan.lifespan(app).__aenter__.await_args_list == [call()]

    @pytest.mark.parametrize(
        ["child_lifespan"],
        (
            pytest.param(MagicMock(), id="lifespan"),
            pytest.param(None, id="no_lifespan"),
        ),
    )
    async def test_shutdown(self, app, lifespan, child_lifespan):
        lifespan.lifespan = child_lifespan
        foo = AsyncMock()
        app.events = MagicMock()
        app.events.shutdown = [foo]

        await lifespan._shutdown(app)

        assert foo.await_args_list == [call()]
        if child_lifespan:
            assert lifespan.lifespan(app).__aexit__.await_args_list == [call(None, None, None)]



import typing as t

from flama import codecs, exceptions, http, routing, schemas, types
from flama.injection import Component, Components
from flama.injection.resolver import Parameter
from flama.negotiation import ContentTypeNegotiator, WebSocketEncodingNegotiator
from flama.schemas import Field, Schema, SchemaValidationError

ValidatedPathParams = t.NewType("ValidatedPathParams", dict)
ValidatedQueryParams = t.NewType("ValidatedQueryParams", dict)
ValidatedRequestData = t.NewType("ValidatedRequestData", dict)


class RequestDataComponent(Component):
    def __init__(self):
        self.negotiator = ContentTypeNegotiator(
            [codecs.JSONDataCodec(), codecs.URLEncodedCodec(), codecs.MultiPartCodec()]
        )

    async def resolve(self, request: http.Request) -> types.RequestData:
        content_type = request.headers.get("Content-Type")

        try:
            codec = self.negotiator.negotiate(content_type)
        except exceptions.NoCodecAvailable:
            raise exceptions.HTTPException(415)

        try:
            data = await codec.decode(request)
            return types.RequestData(data) if data else None  # type: ignore
        except exceptions.DecodeError as exc:
            raise exceptions.HTTPException(400, detail=str(exc))


class WebSocketMessageDataComponent(Component):
    def __init__(self):
        self.negotiator = WebSocketEncodingNegotiator([codecs.BytesCodec(), codecs.TextCodec(), codecs.JSONCodec()])

    async def resolve(self, message: types.Message, websocket_encoding: types.Encoding) -> types.Data:
        try:
            codec = self.negotiator.negotiate(websocket_encoding)
            return types.Data(await codec.decode(message))
        except (exceptions.NoCodecAvailable, exceptions.DecodeError):
            raise exceptions.WebSocketException(code=1003)


class ValidatePathParamsComponent(Component):
    async def resolve(
        self, request: http.Request, route: routing.BaseRoute, path_params: types.PathParams
    ) -> ValidatedPathParams:
        fields = [f.field for f in route.parameters.path[request.method].values()]

        try:
            validated = Schema.build(name="ValidationSchema", fields=fields).validate(path_params)
            return ValidatedPathParams({k: v for k, v in path_params.items() if k in validated})
        except SchemaValidationError as exc:
            raise exceptions.ValidationError(detail=exc.errors)


class ValidateQueryParamsComponent(Component):
    def resolve(
        self, request: http.Request, route: routing.BaseRoute, query_params: types.QueryParams
    ) -> ValidatedQueryParams:
        fields = [f.field for f in route.parameters.query[request.method].values()]

        try:
            validated = Schema.build(name="ValidationSchema", fields=fields).validate(dict(query_params))
            return ValidatedQueryParams({k: v for k, v in query_params.items() if k in validated})
        except SchemaValidationError as exc:
            raise exceptions.ValidationError(detail=exc.errors)


class ValidateRequestDataComponent(Component):
    def resolve(self, request: http.Request, route: routing.BaseRoute, data: types.RequestData) -> ValidatedRequestData:
        body_param = route.parameters.body[request.method]

        assert body_param is not None, (
            f"Body schema parameter not defined for route '{route}' and method '{request.method}'"
        )

        try:
            return ValidatedRequestData(body_param.schema.validate(dict(data)))
        except SchemaValidationError as exc:  # pragma: no cover # safety net, just should not happen
            raise exceptions.ValidationError(detail=exc.errors)


class PrimitiveParamComponent(Component):
    def can_handle_parameter(self, parameter: Parameter):
        return Field.is_http_valid_type(parameter.annotation)

    def resolve(self, parameter: Parameter, path_params: ValidatedPathParams, query_params: ValidatedQueryParams):
        params = path_params if (parameter.name in path_params) else query_params

        try:
            params = Schema.build(name="ValidationSchema", fields=[Field.from_parameter(parameter)]).validate(params)
        except SchemaValidationError as exc:  # pragma: no cover # safety net, just should not happen
            raise exceptions.ValidationError(detail=exc.errors)
        return params.get(parameter.name, parameter.default)


class CompositeParamComponent(Component):
    def can_handle_parameter(self, parameter: Parameter):
        schema = (
            t.get_args(parameter.annotation)[0] if t.get_origin(parameter.annotation) == list else parameter.annotation
        )
        return schemas.is_schema(schema)

    def resolve(self, parameter: Parameter, request: http.Request, route: routing.BaseRoute, data: types.RequestData):
        body_param = route.parameters.body[request.method]

        assert body_param is not None, (
            f"Body schema parameter not defined for route '{route}' and method '{request.method}'"
        )

        try:
            return body_param.schema.validate(data, partial=schemas.is_schema_partial(parameter.annotation))
        except SchemaValidationError as exc:  # pragma: no cover # safety net, just should not happen
            raise exceptions.ValidationError(detail=exc.errors)


VALIDATION_COMPONENTS = Components(
    [
        RequestDataComponent(),
        WebSocketMessageDataComponent(),
        ValidatePathParamsComponent(),
        ValidateQueryParamsComponent(),
        ValidateRequestDataComponent(),
        PrimitiveParamComponent(),
        CompositeParamComponent(),
    ]
)



import dataclasses
import functools
import json
import os
import tempfile
import typing as t

import pytest

from flama.config import Config, exceptions

EMPTY = t.NewType("EMPTY", str)


@dataclasses.dataclass
class Foo:
    bar: int


class TestCaseConfig:
    @pytest.fixture(scope="function")
    def config_file(self, request):
        with tempfile.NamedTemporaryFile("w+") as f:
            json.dump(request.param, f)
            f.flush()
            yield f.name

    @pytest.fixture(scope="function")
    def environment(self, request):
        os.environ.update(request.param)

        yield

        for k in request.param:
            del os.environ[k]

    def test_init_config_file(self):
        with tempfile.NamedTemporaryFile("w+") as f:
            json.dump({"foo": "bar"}, f)
            f.flush()
            config = Config(f.name, "json")

        assert config.config_file == {"foo": "bar"}

    def test_init_no_config_file(self):
        config = Config()
        assert config.config_file == {}

    def test_init_config_file_not_found(self):
        config = Config("not_found.json", "json")
        assert config.config_file == {}

    @pytest.mark.parametrize(
        ["config_file", "environment", "key", "default", "cast", "result", "exception"],
        (
            pytest.param(
                {"foo": "1"},
                {},
                "foo",
                EMPTY,
                EMPTY,
                "1",
                None,
                id="file",
            ),
            pytest.param(
                {},
                {"foo": "1"},
                "foo",
                EMPTY,
                EMPTY,
                "1",
                None,
                id="environment",
            ),
            pytest.param(
                {"foo": "1"},
                {"foo": "2"},
                "foo",
                EMPTY,
                EMPTY,
                "2",
                None,
                id="environment_and_file",
            ),
            pytest.param(
                {},
                {},
                "foo",
                "default",
                EMPTY,
                "default",
                None,
                id="default",
            ),
            pytest.param(
                {"foo": "1"},
                {},
                "foo",
                EMPTY,
                int,
                1,
                None,
                id="cast_type",
            ),
            pytest.param(
                {"foo": "bar"},
                {},
                "foo",
                EMPTY,
                lambda _: 1,
                1,
                None,
                id="cast_function",
            ),
            pytest.param(
                {"foo": "bar"},
                {},
                "foo",
                EMPTY,
                lambda x: int(x),
                None,
                exceptions.ConfigError("Cannot cast config type"),
                id="cast_function_error",
            ),
            pytest.param(
                {"foo": '{"bar": 1}'},
                {},
                "foo",
                EMPTY,
                Foo,
                Foo(bar=1),
                None,
                id="cast_dataclass_str",
            ),
            pytest.param(
                {"foo": '{"bar": 1, "foobar": "foobar"}'},
                {},
                "foo",
                EMPTY,
                Foo,
                Foo(bar=1),
                None,
                id="cast_dataclass_str_subset_of_keys",
            ),
            pytest.param(
                {"foo": {"bar": 1}},
                {},
                "foo",
                EMPTY,
                Foo,
                Foo(bar=1),
                None,
                id="cast_dataclass_dict",
            ),
            pytest.param(
                {"foo": {"bar": 1, "foobar": "foobar"}},
                {},
                "foo",
                EMPTY,
                Foo,
                Foo(bar=1),
                None,
                id="cast_dataclass_dict_subset_of_keys",
            ),
            pytest.param(
                {},
                {},
                "foo",
                '{"bar": 1}',
                Foo,
                Foo(bar=1),
                None,
                id="cast_dataclass_default",
            ),
            pytest.param(
                {"foo": "{wrong_data"},
                {},
                "foo",
                EMPTY,
                Foo,
                None,
                exceptions.ConfigError("Cannot parse value as json for config dataclass"),
                id="cast_dataclass_cannot_parse",
            ),
            pytest.param(
                {"foo": "[1, 2]"},
                {},
                "foo",
                EMPTY,
                Foo,
                None,
                exceptions.ConfigError("Wrong value for config dataclass"),
                id="cast_dataclass_wrong_value",
            ),
            pytest.param(
                {"foo": '{"wrong_key": 1}'},
                {},
                "foo",
                EMPTY,
                Foo,
                None,
                exceptions.ConfigError("Cannot create config dataclass"),
                id="cast_dataclass_no_match_with_dataclass",
            ),
            pytest.param({}, {}, "foo", EMPTY, EMPTY, None, KeyError("foo"), id="not_found"),
        ),
        indirect=["config_file", "environment", "exception"],
    )
    def test_call(self, config_file, environment, key, default, cast, result, exception):
        config = Config(config_file, "json")

        if default is not EMPTY:
            config = functools.partial(config, default=default)

        if cast is not EMPTY:
            config = functools.partial(config, cast=cast)

        with exception:
            assert config(key) == result



import time
import uuid
from unittest.mock import patch

import pytest

from flama.authentication import exceptions
from flama.authentication.jwt.jws import JWS
from flama.authentication.jwt.jwt import JWT

TOKEN = (
    b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V"
    b"28rM_b1UU62XCp9dg_5kg="
)


@pytest.fixture(scope="function")
def key():
    return uuid.UUID(int=0).bytes


class TestCaseJWS:
    @pytest.mark.parametrize(
        ["header", "payload", "result", "exception"],
        (
            pytest.param(
                {"alg": "HS256", "typ": "JWT"},
                {"data": {"foo": "bar"}, "iat": 0},
                TOKEN,
                None,
                id="ok",
            ),
            pytest.param(
                {"alg": "wrong"},
                {},
                None,
                exceptions.JWTDecodeException("Unsupported algorithm 'wrong'"),
                id="unknown_algorithm",
            ),
            pytest.param(
                {},
                {},
                None,
                exceptions.JWTDecodeException("Missing algorithm in header"),
                id="missing_algorithm",
            ),
        ),
        indirect=["exception"],
    )
    def test_encode(self, key, header, payload, result, exception):
        with exception:
            assert JWS.encode(header=header, payload=payload, key=key) == result

    @pytest.mark.parametrize(
        ["token", "result", "exception"],
        (
            pytest.param(
                TOKEN,
                (
                    {"alg": "HS256", "typ": "JWT"},
                    {"data": {"foo": "bar"}, "iat": 0},
                    b"J3zdedMZSFNOimstjJat0V28rM_b1UU62XCp9dg_5kg=",
                ),
                None,
                id="ok",
            ),
            pytest.param(
                b"wrong.format",
                None,
                exceptions.JWTDecodeException("Not enough segments"),
                id="wrong-format",
            ),
            pytest.param(
                b"wrong.format.0000",
                None,
                exceptions.JWTDecodeException("Wrong header format"),
                id="wrong-header",
            ),
            pytest.param(
                b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.format.0000",
                None,
                exceptions.JWTDecodeException("Wrong payload format"),
                id="wrong-payload",
            ),
            pytest.param(
                b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.0000",
                None,
                exceptions.JWTValidateException(
                    "Signature verification failed for token 'eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9."
                    "eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.0000'"
                ),
                id="invalid-signature",
            ),
        ),
        indirect=["exception"],
    )
    def test_decode(self, key, token, result, exception):
        with exception:
            assert JWS.decode(token, key) == result


class TestCaseJWT:
    @pytest.mark.parametrize(
        ["header", "payload", "result", "exception"],
        (
            pytest.param(
                {"alg": "HS256", "typ": "JWT"},
                {"foo": "bar", "iat": 0},
                TOKEN,
                None,
                id="ok",
            ),
            pytest.param(
                {"alg": "wrong"},
                {"foo": "bar"},
                None,
                exceptions.JWTDecodeException("Unsupported algorithm 'wrong'"),
                id="unknown_algorithm",
            ),
            pytest.param(
                {},
                {"foo": "bar"},
                None,
                exceptions.JWTDecodeException("Missing algorithm in header"),
                id="missing_algorithm",
            ),
        ),
        indirect=["exception"],
    )
    def test_encode(self, key, header, payload, result, exception):
        with exception:
            jwt = JWT(header=header, payload=payload).encode(key)
            assert jwt == result

    @pytest.mark.parametrize(
        ["token", "exception"],
        (
            pytest.param(
                JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"foo": "bar", "iat": 0}),
                None,
                id="ok",
            ),
            pytest.param(
                JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"foo": "bar", "iat": time.time() * 2}),
                exceptions.JWTValidateException("Invalid claims (iat)"),
                id="invalid_iat",
            ),
            pytest.param(
                JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"foo": "bar", "exp": time.time() / 2}),
                exceptions.JWTValidateException("Invalid claims (exp)"),
                id="invalid_exp",
            ),
            pytest.param(
                JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"foo": "bar", "nbf": time.time() * 2}),
                exceptions.JWTValidateException("Invalid claims (nbf)"),
                id="invalid_nbf",
            ),
        ),
        indirect=["exception"],
    )
    def test_validate(self, token, exception):
        with exception:
            token.validate()

    @pytest.mark.parametrize(
        ["token", "result", "validate_side_effect", "exception"],
        (
            pytest.param(
                TOKEN,
                JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"data": {"foo": "bar"}, "iat": 0}),
                None,
                None,
                id="ok",
            ),
            pytest.param(
                b"wrong.format",
                None,
                None,
                exceptions.JWTDecodeException("Not enough segments"),
                id="wrong-format",
            ),
            pytest.param(
                b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.0000",
                None,
                None,
                exceptions.JWTValidateException(
                    "Signature verification failed for token 'eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9."
                    "eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.0000'"
                ),
                id="invalid-signature",
            ),
            pytest.param(
                TOKEN,
                JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"data": {"foo": "bar"}, "iat": 0}),
                exceptions.JWTClaimValidateException("exp"),
                exceptions.JWTValidateException("Claim 'exp' is not valid"),
                id="invalid-claims",
            ),
        ),
        indirect=["exception"],
    )
    def test_decode(self, key, token, validate_side_effect, result, exception):
        with exception:
            with patch.object(JWT, "validate", side_effect=validate_side_effect):
                assert JWT.decode(token, key) == result

    def test_to_dict(self):
        jwt = JWT(header={"alg": "HS256", "typ": "JWT"}, payload={"data": {"foo": "bar"}, "iat": 0})

        assert jwt.to_dict() == {
            "header": {"alg": "HS256", "typ": "JWT"},
            "payload": {"data": {"foo": "bar"}, "iat": 0},
        }



import pytest

from flama.config import fields


class TestCaseSecret:
    def test_secret(self):
        secret = fields.Secret("foo")

        assert secret._value == "foo"
        assert repr(secret) == "Secret('*****')"
        assert str(secret) == "foo"
        assert secret == "foo"
        assert secret


class TestCaseURL:
    @pytest.mark.parametrize(
        ["url", "components"],
        (
            pytest.param(
                "",
                {"scheme": "", "netloc": "", "path": "", "params": "", "query": "", "fragment": ""},
                id="empty",
            ),
            pytest.param(
                "https://www.foo.bar/foobar?foo=bar",
                {
                    "scheme": "https",
                    "netloc": "www.foo.bar",
                    "path": "/foobar",
                    "params": "",
                    "query": "foo=bar",
                    "fragment": "",
                },
                id="basic",
            ),
            pytest.param(
                "https://user:pass@www.foo.bar:8000/foobar?foo=bar#barfoo",
                {
                    "scheme": "https",
                    "netloc": "user:pass@www.foo.bar:8000",
                    "path": "/foobar",
                    "params": "",
                    "query": "foo=bar",
                    "fragment": "barfoo",
                },
                id="full",
            ),
        ),
    )
    def test_url(self, url, components):
        url_obj = fields.URL(url)

        assert url_obj.components == components
        assert str(url_obj) == url
        assert repr(url_obj) == f"URL('{url}')"
        assert url_obj == url
        assert url_obj == fields.URL(url)



import typing as t

import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields

from flama import endpoints, http, schemas, types


class TestCaseReturnValidation:
    @pytest.fixture(scope="function")
    def output_schema(self, app):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("OutputSchema", name=(str, ...))
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(title="OutputSchema", fields={"name": typesystem.fields.String()})
        elif app.schema.schema_library.lib == marshmallow:
            schema = type("OutputSchema", (marshmallow.Schema,), {"name": marshmallow.fields.String()})
        else:
            raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")

        return schema

    @pytest.fixture(  # noqa: C901
        scope="function",
        params=[pytest.param(True, id="endpoints"), pytest.param(False, id="function views")],
        autouse=True,
    )
    def add_endpoints(self, app, request, output_schema):  # noqa: C901
        if request.param:

            @app.route("/return-string/")
            class StrEndpoint(endpoints.HTTPEndpoint):
                def get(self, data: types.RequestData) -> str:
                    return "example content"

            @app.route("/return-dict/")
            class DictEndpoint(endpoints.HTTPEndpoint):
                def get(self, data: types.RequestData) -> dict:
                    return {"example": "content"}

            @app.route("/return-html-response/")
            class HTMLEndpoint(endpoints.HTTPEndpoint):
                def get(self, data: types.RequestData) -> http.HTMLResponse:
                    return http.HTMLResponse("<html><body>example content</body></html>")

            @app.route("/return-json-response/")
            class JSONEndpoint(endpoints.HTTPEndpoint):
                def get(self, data: types.RequestData) -> http.JSONResponse:
                    return http.JSONResponse({"example": "content"})

            @app.route("/return-unserializable-json/")
            class UnserializableEndpoint(endpoints.HTTPEndpoint):
                def get(self) -> dict:
                    class Dummy:
                        pass

                    return {"dummy": Dummy()}

            @app.route("/return-schema/", methods=["GET"])
            class ReturnSchemaHTTPEndpoint(endpoints.HTTPEndpoint):
                async def get(self) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(output_schema)]:
                    return {"name": "Canna"}

            @app.route("/return-schema-many/", methods=["GET"])
            class ReturnSchemaManyHTTPEndpoint(endpoints.HTTPEndpoint):
                async def get(self) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(output_schema)]:
                    return [{"name": "Canna"}, {"name": "Sandy"}]

            @app.route("/return-schema-empty/", methods=["GET"])
            class ReturnSchemaEmptyHTTPEndpoint(endpoints.HTTPEndpoint):
                async def get(self) -> None:
                    return None

        else:

            @app.route("/return-string/")
            def return_string(data: types.RequestData) -> str:
                return "example content"

            @app.route("/return-dict/")
            def return_data(data: types.RequestData) -> dict:
                return {"example": "content"}

            @app.route("/return-html-response/")
            def return_html(data: types.RequestData) -> http.HTMLResponse:
                return http.HTMLResponse("<html><body>example content</body></html>")

            @app.route("/return-json-response/")
            def return_response(data: types.RequestData) -> http.Response:
                return http.JSONResponse({"example": "content"})

            @app.route("/return-unserializable-json/")
            def return_unserializable_json() -> dict:
                class Dummy:
                    pass

                return {"dummy": Dummy()}

            @app.route("/return-schema/", methods=["GET"])
            async def return_schema() -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(output_schema)]:
                return {"name": "Canna"}

            @app.route("/return-schema-many/", methods=["GET"])
            async def return_schema_many() -> t.Annotated[
                list[schemas.SchemaType], schemas.SchemaMetadata(output_schema)
            ]:
                return [{"name": "Canna"}, {"name": "Sandy"}]

            @app.route("/return-schema-empty/", methods=["GET"])
            async def return_schema_empty() -> None:
                return None

    @pytest.mark.parametrize(
        ["path", "status_code", "content_type", "content", "exception"],
        (
            pytest.param("/return-string/", 200, "application/json", b'"example content"', None, id="string"),
            pytest.param("/return-dict/", 200, "application/json", b'{"example":"content"}', None, id="dict"),
            pytest.param(
                "/return-html-response/",
                200,
                "text/html; charset=utf-8",
                b"<html><body>example content</body></html>",
                None,
                id="html_response",
            ),
            pytest.param(
                "/return-json-response/", 200, "application/json", b'{"example":"content"}', None, id="json_response"
            ),
            pytest.param(
                "/return-unserializable-json/",
                None,
                None,
                None,
                TypeError("Object of type Dummy is not JSON serializable"),
                id="unserializable_json_response",
            ),
            pytest.param("/return-schema/", 200, "application/json", b'{"name":"Canna"}', None, id="schema"),
            pytest.param(
                "/return-schema-many/",
                200,
                "application/json",
                b'[{"name":"Canna"},{"name":"Sandy"}]',
                None,
                id="schema_many",
            ),
            pytest.param("/return-schema-empty/", 200, "application/json", b"", None, id="schema_empty"),
        ),
        indirect=["exception"],
    )
    async def test_return(self, client, path, status_code, content_type, content, exception):
        with exception:
            response = await client.get(path)

            assert response.status_code == status_code
            assert response.headers["content-type"] == content_type
            assert response.content == content



import os
import typing as t

import flama.schemas
from flama import schemas
from flama.models.components import ModelComponentBuilder
from flama.resources import data_structures
from flama.resources.exceptions import ResourceAttributeError
from flama.resources.resource import Resource, ResourceType
from flama.resources.routing import resource_method

if t.TYPE_CHECKING:
    from flama.models.base import Model
    from flama.models.components import ModelComponent

__all__ = ["BaseModelResource", "ModelResource", "InspectMixin", "PredictMixin", "ModelResourceType"]


Component = t.TypeVar("Component", bound="ModelComponent")


class InspectMixin:
    @classmethod
    def _add_inspect(cls, name: str, verbose_name: str, model_model_type: type["Model"], **kwargs) -> dict[str, t.Any]:
        @resource_method("/", methods=["GET"], name="inspect")
        async def inspect(self, model: model_model_type):  # type: ignore[valid-type]
            return model.inspect()  # type: ignore[attr-defined]

        inspect.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Retrieve the model
            description:
                Retrieve the model from this resource.
            responses:
                200:
                    description:
                        The model.
        """

        return {"_inspect": inspect}


class PredictMixin:
    @classmethod
    def _add_predict(cls, name: str, verbose_name: str, model_model_type: type["Model"], **kwargs) -> dict[str, t.Any]:
        @resource_method("/predict/", methods=["POST"], name="predict")
        async def predict(
            self,
            model: model_model_type,  # type: ignore[valid-type]
            data: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(flama.schemas.schemas.MLModelInput)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(flama.schemas.schemas.MLModelOutput)]:
            return {"output": model.predict(data["input"])}

        predict.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Generate a prediction
            description:
                Generate a prediction using the model from this resource.
            responses:
                200:
                    description:
                        The prediction generated by the model.
        """

        return {"_predict": predict}


class ModelResourceType(ResourceType, InspectMixin, PredictMixin):
    METHODS = ("inspect", "predict")

    def __new__(mcs, name: str, bases: tuple[type], namespace: dict[str, t.Any]):
        """Resource metaclass for defining basic behavior for ML resources:
        * Create _meta attribute containing some metadata (model...).
        * Adds methods related to ML resource (inspect, predict...) listed in METHODS class attribute.

        :param name: Class name.
        :param bases: List of superclasses.
        :param namespace: Variables namespace used to create the class.
        """
        if not mcs._is_abstract(namespace):
            try:
                # Get model component
                component = mcs._get_model_component(bases, namespace)
                namespace["component"] = component
                namespace["model"] = component.model
            except AttributeError as e:
                raise ResourceAttributeError(str(e), name)

            namespace.setdefault("_meta", data_structures.Metadata()).namespaces["model"] = {
                "component": component,
                "model": component.model,
                "model_type": component.get_model_type(),
            }

        return super().__new__(mcs, name, bases, namespace)

    @staticmethod
    def _is_abstract(namespace: dict[str, t.Any]) -> bool:
        return namespace.get("__module__") == "flama.models.resource" and namespace.get("__qualname__") in (
            "BaseModelResource",
            "ModelResource",
        )

    @classmethod
    def _get_model_component(cls, bases: t.Sequence[t.Any], namespace: dict[str, t.Any]) -> "ModelComponent":
        try:
            component: ModelComponent = cls._get_attribute("component", bases, namespace, metadata_namespace="model")
            return component
        except AttributeError:
            ...

        try:
            return ModelComponentBuilder.load(
                cls._get_attribute("model_path", bases, namespace, metadata_namespace="model")
            )
        except AttributeError:
            ...

        raise AttributeError(ResourceAttributeError.MODEL_NOT_FOUND)


class BaseModelResource(Resource, t.Generic[Component], metaclass=ModelResourceType):
    component: Component
    model: t.Any
    model_path: t.Union[str, os.PathLike]


class ModelResource(BaseModelResource["ModelComponent"]): ...



import dataclasses
import inspect
import sys
import typing as t
from pathlib import Path

if t.TYPE_CHECKING:
    from flama import http
    from flama.applications import Flama
    from flama.routing import Route, WebSocketRoute


@dataclasses.dataclass(frozen=True)
class RequestParams:
    path: dict[str, t.Any]
    query: dict[str, t.Any]


@dataclasses.dataclass(frozen=True)
class RequestClient:
    host: str
    port: int


@dataclasses.dataclass(frozen=True)
class Request:
    path: str
    method: str
    params: RequestParams
    headers: dict[str, str]
    cookies: dict[str, str]
    client: t.Optional[RequestClient] = None

    @classmethod
    def from_request(cls, request: "http.Request") -> "Request":
        return cls(
            path=request.url.path,
            method=request.method,
            params=RequestParams(query=dict(request.query_params), path=dict(request.path_params)),
            headers=dict(request.headers),
            cookies=dict(request.cookies),
            client=RequestClient(host=request.client.host, port=request.client.port) if request.client else None,
        )


@dataclasses.dataclass(frozen=True)
class Frame:
    filename: str
    function: str
    line: int
    vendor: t.Optional[str]
    code: str

    @classmethod
    def from_frame_info(cls, frame: inspect.FrameInfo) -> "Frame":
        filename = Path(frame.filename).resolve()
        for sys_path in sys.path:
            try:
                relative_filename = filename.relative_to(sys_path)
                break
            except ValueError:  # pragma: no cover # Just a safety mechanism
                ...
        else:  # pragma: no cover # Just a safety mechanism
            raise ValueError(f"Filename '{filename}' not found in sys.path")

        try:
            vendor = str(filename.relative_to(sys.exec_prefix).parts[3])
        except ValueError:
            vendor = None

        with open(filename) as f:
            code = f.read()

        return cls(
            filename=str(relative_filename),
            function=frame.function,
            line=frame.lineno,
            vendor=vendor,
            code=code,
        )


@dataclasses.dataclass(frozen=True)
class Error:
    error: str
    description: str
    traceback: list[Frame]

    @classmethod
    def from_exception(cls, exc: Exception, context: int = 10) -> "Error":
        frames = inspect.getinnerframes(exc.__traceback__, context) if exc.__traceback__ else []
        exc_cls = exc if inspect.isclass(exc) else exc.__class__
        return cls(
            error=(
                f"{exc_cls.__module__}.{exc_cls.__name__}"
                if exc_cls.__module__ not in ("builtins", "__main__")
                else exc_cls.__name__
            ),
            description=str(exc),
            traceback=[Frame.from_frame_info(frame=frame) for frame in frames],
        )


@dataclasses.dataclass(frozen=True)
class Environment:
    platform: str
    python: str
    python_version: str
    path: list[str]

    @classmethod
    def from_system(cls) -> "Environment":
        return cls(platform=sys.platform, python=sys.executable, python_version=sys.version, path=sys.path)


@dataclasses.dataclass(frozen=True)
class Endpoint:
    path: str
    endpoint: str
    module: t.Optional[str]
    file: str
    line: int
    name: t.Optional[str] = None

    @classmethod
    def from_route(cls, route: t.Union["Route", "WebSocketRoute"]) -> "Endpoint":
        handler = route.app.handler
        module = inspect.getmodule(route.app.handler)
        filename = Path(inspect.getfile(route.app.handler)).resolve()
        for sys_path in sys.path:
            try:
                relative_filename = filename.relative_to(sys_path)
                break
            except ValueError:  # pragma: no cover # Just a safety mechanism
                ...
        else:  # pragma: no cover # Just a safety mechanism
            raise ValueError(f"Filename '{filename}' not found in sys.path")

        return cls(
            path=str(route.path),
            endpoint=handler.__name__,
            module=module.__name__ if module else None,
            file=str(relative_filename),
            line=inspect.getsourcelines(route.app.handler)[1],
            name=route.name,
        )


@dataclasses.dataclass(frozen=True)
class App:
    apps: list["App"]
    endpoints: list[Endpoint]
    path: str
    name: t.Optional[str] = None

    @classmethod
    def from_app(cls, app: t.Any, path: str = "/", name: t.Optional[str] = None) -> "App":
        apps: list[App] = []
        endpoints: list[Endpoint] = []
        for route in app.routes:
            try:
                apps.append(App.from_app(route.app, path=route.path.path, name=route.name))
            except AttributeError:
                endpoints.append(Endpoint.from_route(route))

        return cls(apps=apps, endpoints=endpoints, path=path, name=name)


@dataclasses.dataclass(frozen=True)
class ErrorContext:
    request: Request
    environment: Environment
    error: Error

    @classmethod
    def build(cls, request: "http.Request", exc: Exception) -> "ErrorContext":
        return cls(
            request=Request.from_request(request),
            environment=Environment.from_system(),
            error=Error.from_exception(exc),
        )


@dataclasses.dataclass(frozen=True)
class NotFoundContext:
    request: Request
    environment: Environment
    app: App

    @classmethod
    def build(cls, request: "http.Request", app: "Flama") -> "NotFoundContext":
        return cls(request=Request.from_request(request), environment=Environment.from_system(), app=App.from_app(app))



import typing as t

import click

from flama.cli.config.app import App
from flama.cli.config.config import Config
from flama.cli.config.config import options as config_options
from flama.cli.config.uvicorn import Uvicorn
from flama.cli.config.uvicorn import options as uvicorn_options

__all__ = ["run", "command"]


@click.command(name="run", context_settings={"auto_envvar_prefix": "FLAMA"})
@click.argument("flama-app", envvar="FLAMA_APP")
@config_options
@uvicorn_options
def command(flama_app: str, uvicorn: Uvicorn):
    """Run a Flama Application based on a route.

    <FLAMA_APP> is the route to the Flama object to be served, e.g. 'examples.hello_flama:app'. This can be passed
    directly as argument of the command line, or by environment variable."""
    Config(app=App.build(flama_app), server=uvicorn).run()


run: t.Callable = command.callback  # type: ignore[assignment]



from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
import starlette.exceptions

from flama import exceptions, http, types, websockets
from flama.applications import Flama
from flama.debug.data_structures import ErrorContext
from flama.debug.middleware import BaseErrorMiddleware, ExceptionMiddleware, ServerErrorMiddleware


class TestCaseBaseErrorMiddleware:
    @pytest.fixture
    def middleware_cls(self):
        class FooMiddleware(BaseErrorMiddleware):
            async def process_exception(
                self,
                scope: types.Scope,
                receive: types.Receive,
                send: types.Send,
                exc: Exception,
                response_started: bool,
            ) -> None: ...

        return FooMiddleware

    def test_init(self, middleware_cls):
        app = AsyncMock()
        middleware = middleware_cls(app=app, debug=True)
        assert middleware.app == app
        assert middleware.debug

    async def test_call_http(self, middleware_cls, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "http"
        exc = ValueError()
        app = AsyncMock(side_effect=exc)
        middleware = middleware_cls(app=app, debug=True)
        with patch.object(middleware, "process_exception", new_callable=AsyncMock):
            await middleware(asgi_scope, asgi_receive, asgi_send)
            assert middleware.app.call_count == 1
            scope, receive, send = middleware.app.call_args_list[0][0]
            assert scope == asgi_scope
            assert receive == asgi_receive
            assert middleware.process_exception.call_args_list == [
                call(asgi_scope, asgi_receive, asgi_send, exc, False)
            ]

    async def test_call_websocket(self, middleware_cls, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "websocket"
        exc = ValueError()
        app = AsyncMock(side_effect=exc)
        middleware = middleware_cls(app=app, debug=True)
        with patch.object(middleware, "process_exception", new_callable=AsyncMock):
            await middleware(asgi_scope, asgi_receive, asgi_send)
            assert middleware.app.call_count == 1
            scope, receive, send = middleware.app.call_args_list[0][0]
            assert scope == asgi_scope
            assert receive == asgi_receive
            assert middleware.process_exception.call_args_list == [
                call(asgi_scope, asgi_receive, asgi_send, exc, False)
            ]


class TestCaseServerErrorMiddleware:
    @pytest.fixture
    def middleware(self):
        return ServerErrorMiddleware(app=AsyncMock())

    @pytest.mark.parametrize(
        ["request_type", "debug", "handler"],
        (
            pytest.param("http", True, "debug_handler", id="http_debug"),
            pytest.param("http", False, "error_handler", id="http_no_debug"),
            pytest.param("websocket", True, "noop_handler", id="websocket_debug"),
            pytest.param("websocket", False, "noop_handler", id="websocket_no_debug"),
        ),
    )
    def test_get_handler(self, middleware, asgi_scope, request_type, debug, handler):
        asgi_scope["type"] = request_type
        middleware.debug = debug

        result_handler = middleware._get_handler(asgi_scope)

        assert result_handler == getattr(middleware, handler)

    @pytest.mark.parametrize(
        ["debug", "response_started"],
        (
            pytest.param(True, False, id="debug_not_started"),
            pytest.param(True, True, id="debug_started"),
            pytest.param(False, False, id="error_not_started"),
            pytest.param(False, True, id="error_started"),
        ),
    )
    async def test_process_exception(self, middleware, asgi_scope, asgi_receive, asgi_send, debug, response_started):
        middleware.debug = debug
        response_method = "debug_handler" if debug else "error_handler"
        exc = ValueError("Foo")
        with (
            patch.object(ServerErrorMiddleware, response_method, new=MagicMock(return_value=AsyncMock())) as response,
            pytest.raises(ValueError, match="Foo"),
        ):
            await middleware.process_exception(asgi_scope, asgi_receive, asgi_send, exc, response_started)

            if debug:
                assert ServerErrorMiddleware.debug_handler.call_args_list == [call(asgi_scope, exc)]
            else:
                assert ServerErrorMiddleware.error_handler.call_args_list == [call(asgi_scope, exc)]

            if response_started:
                assert response.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]
            else:
                assert response.call_args_list == []

    def test_debug_response_html(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["headers"].append((b"accept", b"text/html"))
        exc = ValueError()
        error_context_mock, context_mock = MagicMock(), MagicMock()
        with (
            patch("flama.debug.middleware.dataclasses.asdict", return_value=context_mock) as dataclasses_dict,
            patch.object(ErrorContext, "build", return_value=error_context_mock),
            patch.object(http._FlamaTemplateResponse, "__init__", return_value=None) as response_mock,
        ):
            response = middleware.debug_handler(asgi_scope, asgi_receive, asgi_send, exc)
            assert ErrorContext.build.call_count == 1
            assert dataclasses_dict.call_args_list == [call(error_context_mock)]
            assert isinstance(response, http._FlamaTemplateResponse)
            assert response_mock.call_args_list == [call("debug/error_500.html", context=context_mock, status_code=500)]

    def test_debug_response_text(self, middleware, asgi_scope, asgi_receive, asgi_send):
        exc = ValueError()
        with patch.object(http.PlainTextResponse, "__init__", return_value=None) as response_mock:
            response = middleware.debug_handler(asgi_scope, asgi_receive, asgi_send, exc)
            assert isinstance(response, http.PlainTextResponse)
            assert response_mock.call_args_list == [call("Internal Server Error", status_code=500)]

    def test_error_response(self, middleware, asgi_scope, asgi_receive, asgi_send):
        exc = ValueError()
        with patch.object(http.PlainTextResponse, "__init__", return_value=None) as response_mock:
            response = middleware.error_handler(asgi_scope, asgi_receive, asgi_send, exc)
            assert isinstance(response, http.PlainTextResponse)
            assert response_mock.call_args_list == [call("Internal Server Error", status_code=500)]


class TestCaseExceptionMiddleware:
    @pytest.fixture
    def middleware(self):
        return ExceptionMiddleware(app=AsyncMock())

    @pytest.fixture
    def handler(self):
        def _handler(): ...

        return _handler

    def test_init(self, handler):
        app = AsyncMock()
        debug = True

        middleware = ExceptionMiddleware(app=app, handlers={400: handler, ValueError: handler}, debug=debug)

        assert middleware.app == app
        assert middleware.debug == debug
        assert middleware._status_handlers == {400: handler}
        assert middleware._exception_handlers == {
            ValueError: handler,
            exceptions.NotFoundException: middleware.not_found_handler,
            exceptions.MethodNotAllowedException: middleware.method_not_allowed_handler,
            starlette.exceptions.HTTPException: middleware.http_exception_handler,
            starlette.exceptions.WebSocketException: middleware.websocket_exception_handler,
        }

    @pytest.mark.parametrize(
        ["status_code", "exc_class", "exception"],
        (
            pytest.param(400, None, None, id="status_code"),
            pytest.param(None, ValueError, None, id="exc_class"),
            pytest.param(400, ValueError, None, id="status_code_and_exc_class"),
            pytest.param(None, None, ValueError("Status code or exception class must be defined"), id="no_key"),
        ),
        indirect=["exception"],
    )
    def test_add_exception_handler(self, middleware, handler, status_code, exc_class, exception):
        status_code_handlers = {}
        if status_code is not None:
            status_code_handlers[status_code] = handler
        exc_class_handlers = {
            exceptions.NotFoundException: middleware.not_found_handler,
            exceptions.MethodNotAllowedException: middleware.method_not_allowed_handler,
            starlette.exceptions.HTTPException: middleware.http_exception_handler,
            starlette.exceptions.WebSocketException: middleware.websocket_exception_handler,
        }
        if exc_class is not None:
            exc_class_handlers[exc_class] = handler

        with exception:
            middleware.add_exception_handler(handler, status_code=status_code, exc_class=exc_class)

        assert middleware._status_handlers == status_code_handlers
        assert middleware._exception_handlers == exc_class_handlers

    @pytest.mark.parametrize(
        ["status_code", "exc_class", "key", "exception"],
        (
            pytest.param(400, None, exceptions.HTTPException(400), None, id="status_code"),
            pytest.param(None, ValueError, ValueError("Foo"), None, id="exc_class"),
            pytest.param(400, ValueError, ValueError("Foo"), None, id="status_code_and_exc_class"),
            pytest.param(None, Exception, ValueError("Foo"), None, id="child_exc_class"),
            pytest.param(
                400, None, exceptions.HTTPException(401), exceptions.HTTPException(401), id="handler_not_found"
            ),
        ),
        indirect=["exception"],
    )
    def test_get_handler(self, middleware, handler, status_code, exc_class, key, exception):
        # Force clean all handlers
        middleware._status_handlers = {}
        middleware._exception_handlers = {}
        middleware.add_exception_handler(handler=handler, status_code=status_code, exc_class=exc_class)

        with exception:
            result_handler = middleware._get_handler(key)
            if not exception:
                assert result_handler == handler

    @pytest.mark.parametrize(
        ["request_type", "response_started", "exception"],
        (
            pytest.param("http", False, None, id="http"),
            pytest.param("websocket", False, None, id="websocket"),
            pytest.param(
                None,
                True,
                RuntimeError("Caught handled exception, but response already started."),
                id="response_started_error",
            ),
        ),
        indirect=["exception"],
    )
    async def test_process_exception(
        self, middleware, asgi_scope, asgi_receive, asgi_send, request_type, response_started, exception
    ):
        expected_exc = ValueError()
        asgi_scope["type"] = request_type
        handler_mock = MagicMock()
        response_mock = AsyncMock() if request_type == "http" else None
        with (
            exception,
            patch.object(middleware, "_get_handler", return_value=handler_mock),
            patch("flama.debug.middleware.concurrency.run", new=AsyncMock(return_value=response_mock)) as run_mock,
        ):
            await middleware.process_exception(asgi_scope, asgi_receive, asgi_send, expected_exc, response_started)

            if request_type == "http":
                assert run_mock.call_count == 1
                handler, scope, receive, send, exc = run_mock.call_args_list[0][0]
                assert handler == handler_mock
                assert scope == asgi_scope
                assert receive == asgi_receive
                assert exc == expected_exc
                assert response_mock.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]

            elif request_type == "websocket":
                assert run_mock.call_count == 1
                handler, scope, receive, send, exc = run_mock.call_args_list[0][0]
                assert handler == handler_mock
                assert scope == asgi_scope
                assert receive == asgi_receive
                assert exc == expected_exc

    @pytest.mark.parametrize(
        ["debug", "accept", "exc", "response_class", "response_params"],
        (
            pytest.param(
                False,
                None,
                exceptions.HTTPException(204),
                http.Response,
                {"status_code": 204, "headers": None},
                id="204",
            ),
            pytest.param(
                False,
                None,
                exceptions.HTTPException(304),
                http.Response,
                {"status_code": 304, "headers": None},
                id="304",
            ),
            pytest.param(
                True,
                b"text/html",
                exceptions.HTTPException(404, "Foo"),
                http._FlamaTemplateResponse,
                {"template": "debug/error_404.html", "context": {}, "status_code": 404},
                id="debug_404",
            ),
            pytest.param(
                False,
                None,
                exceptions.HTTPException(400, "Foo"),
                http.APIErrorResponse,
                {"detail": "Foo", "status_code": 400},
                id="other",
            ),
        ),
    )
    def test_http_exception_handler(
        self, middleware, asgi_scope, asgi_receive, asgi_send, debug, accept, exc, response_class, response_params
    ):
        asgi_scope["type"] = "http"
        asgi_scope["app"] = MagicMock(Flama)
        middleware.debug = debug

        if accept:
            asgi_scope["headers"].append((b"accept", accept))

        if response_class == http.APIErrorResponse:
            response_params["exception"] = exc

        with (
            patch(f"flama.debug.middleware.http.{response_class.__name__}", spec=response_class) as response_mock,
            patch("flama.debug.middleware.dataclasses.asdict", return_value={}),
        ):
            response = middleware.http_exception_handler(asgi_scope, asgi_receive, asgi_send, exc)

            assert isinstance(response, response_class)
            assert response_mock.call_args_list == [call(**response_params)]

    async def test_websocket_exception_handler(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "websocket"
        exc = exceptions.WebSocketException(1011, "Foo reason")

        with patch.object(websockets.WebSocket, "close", new_callable=AsyncMock) as websocket_close_mock:
            await middleware.websocket_exception_handler(asgi_scope, asgi_receive, asgi_send, exc)

            assert websocket_close_mock.call_args_list == [call(code=exc.code, reason=exc.reason)]

    async def test_not_found_handler_http(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "http"
        asgi_scope["app"] = MagicMock(Flama)
        exc = exceptions.NotFoundException()

        with patch("flama.debug.middleware.ExceptionMiddleware.http_exception_handler") as mock:
            await middleware.not_found_handler(asgi_scope, asgi_receive, asgi_send, exc)

        assert mock.call_args_list == [
            call(asgi_scope, asgi_receive, asgi_send, exc=exceptions.HTTPException(status_code=404))
        ]

    async def test_not_found_handler_http_no_app(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "http"
        del asgi_scope["app"]
        exc = exceptions.NotFoundException()

        with patch("flama.debug.middleware.http.PlainTextResponse") as mock:
            await middleware.not_found_handler(asgi_scope, asgi_receive, asgi_send, exc)

        assert mock.call_args_list == [call("Not Found", status_code=404)]

    async def test_not_found_handler_websocket(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "websocket"
        exc = exceptions.NotFoundException()

        with patch("flama.debug.middleware.ExceptionMiddleware.websocket_exception_handler") as mock:
            await middleware.not_found_handler(asgi_scope, asgi_receive, asgi_send, exc)

        assert mock.call_args_list == [
            call(asgi_scope, asgi_receive, asgi_send, exc=exceptions.WebSocketException(1000))
        ]

    async def test_method_not_allowed_handler_http(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "http"
        asgi_scope["app"] = MagicMock(Flama)
        exc = exceptions.MethodNotAllowedException("/", "POST", {"GET"})

        with patch("flama.debug.middleware.ExceptionMiddleware.http_exception_handler") as mock:
            await middleware.method_not_allowed_handler(asgi_scope, asgi_receive, asgi_send, exc)

        assert mock.call_args_list == [
            call(
                asgi_scope,
                asgi_receive,
                asgi_send,
                exc=exceptions.HTTPException(status_code=405, headers={"Allow": "GET"}),
            )
        ]

    async def test_method_not_allowed_handler_websocket(self, middleware, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "websocket"
        exc = exceptions.MethodNotAllowedException("/", "POST", {"GET"})

        with patch("flama.debug.middleware.ExceptionMiddleware.websocket_exception_handler") as mock:
            await middleware.method_not_allowed_handler(asgi_scope, asgi_receive, asgi_send, exc)

        assert mock.call_args_list == [
            call(asgi_scope, asgi_receive, asgi_send, exc=exceptions.WebSocketException(1000))
        ]



import pytest

from flama.models import ModelResource, ModelResourceType
from flama.resources.exceptions import ResourceAttributeError


class TestCaseModelResource:
    def test_resource_using_component(self, app, model, component):
        component_ = component

        @app.models.model_resource("/")
        class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
            name = "puppy"
            verbose_name = "Puppy"
            component = component_

        resource = PuppyModelResource()

        assert not hasattr(resource, "name")
        assert not hasattr(resource, "verbose_name")
        assert hasattr(resource, "component")
        assert resource.component == component
        assert hasattr(resource, "model")
        assert resource.model == model
        assert hasattr(resource, "_meta")
        assert resource._meta.name == "puppy"
        assert resource._meta.verbose_name == "Puppy"
        assert resource._meta.namespaces == {
            "model": {"component": component, "model": model, "model_type": component.get_model_type()}
        }

    @pytest.mark.parametrize(
        ["model_path"],
        (
            pytest.param("sklearn", id="sklearn"),
            pytest.param("sklearn-pipeline", id="sklearn-pipeline"),
            pytest.param("tensorflow", id="tensorflow"),
            pytest.param("torch", id="torch"),
        ),
        indirect=["model_path"],
    )
    def test_resource_using_model_path(self, app, model_path):
        model_path_ = model_path

        class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
            name = "puppy"
            verbose_name = "Puppy"
            model_path = model_path_

        resource = PuppyModelResource()

        app.models.add_model_resource("/", resource)

        assert not hasattr(resource, "name")
        assert not hasattr(resource, "verbose_name")
        assert hasattr(resource, "component")
        component = resource.component
        assert hasattr(resource, "model")
        assert resource.model == component.model
        assert hasattr(resource, "_meta")
        assert resource._meta.name == "puppy"
        assert resource._meta.verbose_name == "Puppy"
        assert resource._meta.namespaces == {
            "model": {"component": component, "model": component.model, "model_type": component.get_model_type()}
        }

    def test_resource_wrong(self):
        with pytest.raises(ResourceAttributeError):

            class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
                name = "puppy"
                verbose_name = "Puppy"


class TestCaseModelResourceMethods:
    @pytest.mark.parametrize(
        ("lib", "model_path", "url"),
        (
            pytest.param("torch", "torch", "/torch/", id="torch"),
            pytest.param("sklearn", "sklearn", "/sklearn/", id="sklearn"),
            pytest.param("sklearn-pipeline", "sklearn-pipeline", "/sklearn-pipeline/", id="sklearn-pipeline"),
            pytest.param("tensorflow", "tensorflow", "/tensorflow/", id="tensorflow"),
        ),
        indirect=["model_path"],
    )
    async def test_inspect(self, app, client, lib, model_path, url):
        app.models.add_model(f"/{lib}/", model=model_path, name=lib)

        response = await client.get(url)
        assert response.status_code == 200, response.json()
        inspect = response.json()
        assert set(inspect.keys()) == {"meta", "artifacts"}
        meta = inspect["meta"]
        assert set(meta.keys()) == {"id", "timestamp", "model", "framework", "extra"}
        assert set(meta["model"].keys()) == {"obj", "info", "params", "metrics"}
        assert set(meta["framework"].keys()) == {"lib", "version"}

    @pytest.mark.parametrize(
        ("lib", "model_path", "url", "x", "y", "status_code"),
        (
            pytest.param(
                "torch",
                "torch",
                "/torch/predict/",
                [[0, 0], [0, 1], [1, 0], [1, 1]],
                [[0], [1], [1], [0]],
                200,
                id="torch-200",
            ),
            pytest.param("torch", "torch", "/torch/predict/", [["wrong"]], None, 400, id="torch-400"),
            pytest.param(
                "sklearn",
                "sklearn",
                "/sklearn/predict/",
                [[0, 0], [0, 1], [1, 0], [1, 1]],
                [0, 1, 1, 0],
                200,
                id="sklearn-200",
            ),
            pytest.param("sklearn", "sklearn", "/sklearn/predict/", [["wrong"]], None, 400, id="sklearn-400"),
            pytest.param(
                "sklearn-pipeline",
                "sklearn-pipeline",
                "/sklearn-pipeline/predict/",
                [[0, 0], [0, 1], [1, 0], [1, 1]],
                [0, 1, 1, 0],
                200,
                id="sklearn-pipeline-200",
            ),
            pytest.param(
                "sklearn-pipeline",
                "sklearn-pipeline",
                "/sklearn-pipeline/predict/",
                [["wrong"]],
                None,
                400,
                id="sklearn-pipeline-400",
            ),
            pytest.param(
                "tensorflow",
                "tensorflow",
                "/tensorflow/predict/",
                [[0, 0], [0, 1], [1, 0], [1, 1]],
                [[0], [1], [1], [0]],
                200,
                id="tensorflow-200",
            ),
            pytest.param(
                "tensorflow", "tensorflow", "/tensorflow/predict/", [["wrong"]], None, 400, id="tensorflow-400"
            ),
        ),
        indirect=["model_path"],
    )
    async def test_predict(self, app, client, lib, model_path, url, x, y, status_code):
        app.models.add_model(f"/{lib}/", model=model_path, name=lib)

        response = await client.post(url, json={"input": x})

        assert response.status_code == status_code, response.json()
        if status_code == 200:
            for a, e in zip(response.json()["output"], y):
                assert a == pytest.approx(e, abs=3e-1)



import abc
import contextlib
import dataclasses
import functools
import itertools
import os
import tempfile
import typing as t
from pathlib import Path

import click
import jinja2

if t.TYPE_CHECKING:
    from flama.applications import Flama

__all__ = ["Model", "App", "options"]

model_decorators = (
    click.argument("model-path", envvar="FLAMA_MODEL_PATH"),
    click.option("--model-url", envvar="MODEL_URL", default="/", show_default=True, help="Route of the model"),
    click.option("--model-name", envvar="MODEL_NAME", default="model", show_default=True, help="Name of the model"),
)

app_decorators = (
    click.option("--app-debug", envvar="APP_DEBUG", is_flag=True, default=False, show_default=True, help="Debug mode"),
    click.option("--app-title", envvar="APP_TITLE", default="Flama", show_default=True, help="Name of the application"),
    click.option(
        "--app-version", envvar="APP_VERSION", default="0.1.0", show_default=True, help="Version of the application"
    ),
    click.option(
        "--app-description",
        envvar="APP_DESCRIPTION",
        default="Fire up with the flame",
        show_default=True,
        help="Description of the application",
    ),
    click.option(
        "--app-schema",
        envvar="APP_SCHEMA",
        default="/schema/",
        show_default=True,
        help="Route of the application schema",
    ),
    click.option(
        "--app-docs",
        envvar="APP_DOCS",
        default="/docs/",
        show_default=True,
        help="Route of the application documentation",
    ),
)


class _AppContext:
    def __init__(self, app: t.Union[str, "Flama"], path: Path, module: t.Optional[str] = None):
        self._app = app
        self._module = module
        self._path = path

    @property
    def dir(self) -> str:
        return str(self._path)

    @property
    def app(self) -> t.Union[str, "Flama"]:
        return f"{self._module}:{self._app}" if isinstance(self._app, str) else self._app


@dataclasses.dataclass
class Model:
    url: str
    path: str
    name: str


class App(metaclass=abc.ABCMeta):
    @property  # type: ignore
    @abc.abstractmethod
    @contextlib.contextmanager
    def context(self) -> t.Generator[_AppContext, None, None]: ...

    @classmethod
    def build(cls, app: t.Union[str, dict[str, t.Any], "Flama"]) -> "App":
        if isinstance(app, str):
            return StrApp(app)

        if isinstance(app, dict):
            return DictApp.from_dict(app)

        return FlamaApp(app)


@dataclasses.dataclass
class FlamaApp(App):
    app: "Flama"

    @property  # type: ignore
    @contextlib.contextmanager
    def context(self) -> t.Generator[_AppContext, None, None]:
        yield _AppContext(app=self.app, path=Path(os.getcwd()))


@dataclasses.dataclass
class DictApp(App):
    debug: bool = False
    title: str = "Flama"
    version: str = "0.1.0"
    description: str = "Fire up with the flame"
    schema: str = "/schema/"
    docs: str = "/docs/"
    models: list[Model] = dataclasses.field(
        default_factory=lambda: [Model(url="/model-url/", path="model-path.flm", name="model-name")]
    )

    @classmethod
    def from_dict(cls, data: dict[str, t.Any]) -> "App":
        if "models" in data:
            data["models"] = [Model(**model) for model in data.pop("models")]
        return cls(**data)  # type: ignore[arg-type]

    @property  # type: ignore
    @contextlib.contextmanager
    def context(self) -> t.Generator[_AppContext, None, None]:
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".py") as f:
            env = jinja2.Environment(loader=jinja2.PackageLoader("flama", "cli/templates"))
            f.write(env.get_template("app.py.j2").render(**dataclasses.asdict(self)))
            f.flush()
            file_path = Path(f.name)
            yield _AppContext(app="app", path=file_path.parent, module=file_path.stem)


class StrApp(str, App):
    @property  # type: ignore
    @contextlib.contextmanager
    def context(self) -> t.Generator[_AppContext, None, None]:
        module, app = self.split(":")
        yield _AppContext(app=app, path=Path(os.getcwd()), module=module)


def options(command: t.Callable) -> t.Callable:
    """Decorate a click command with all App options.

    :param command: Command to be decorated.
    :return: Decorated command.
    """

    @functools.wraps(command)
    def _inner(
        app_debug: bool,
        app_title: str,
        app_version: str,
        app_description: str,
        app_schema: str,
        app_docs: str,
        model_path: str,
        model_url: str,
        model_name: str,
        *args,
        **kwargs,
    ):
        command(
            app=DictApp(
                debug=app_debug,
                title=app_title,
                description=app_description,
                version=app_version,
                schema=app_schema,
                docs=app_docs,
                models=[Model(path=model_path, url=model_url, name=model_name)],
            ),
            *args,
            **kwargs,
        )

    return functools.reduce(lambda x, y: y(x), itertools.chain(app_decorators[::-1], model_decorators[::-1]), _inner)



import http
import logging
import typing as t

from flama import Component
from flama.authentication import exceptions, jwt, types
from flama.exceptions import HTTPException
from flama.types import Headers
from flama.types.http import Cookies

logger = logging.getLogger(__name__)

__all__ = ["AccessTokenComponent", "RefreshTokenComponent"]


class BaseTokenComponent(Component):
    def __init__(self, secret: bytes, *, header_key: str, header_prefix: str, cookie_key: str):
        self.secret = secret
        self.header_key = header_key
        self.header_prefix = header_prefix
        self.cookie_key = cookie_key

    def _token_from_cookies(self, cookies: Cookies) -> bytes:
        try:
            token = cookies[self.cookie_key]["value"]
        except KeyError:
            logger.debug("'%s' not found in cookies", self.cookie_key)
            raise exceptions.Unauthorized()

        return token.encode()

    def _token_from_header(self, headers: Headers) -> bytes:
        try:
            header_prefix, token = headers[self.header_key].split()
        except KeyError:
            logger.debug("'%s' not found in headers", self.header_key)
            raise exceptions.Unauthorized()
        except ValueError:
            logger.debug("Wrong format for authorization header value")
            raise exceptions.JWTException(
                f"Authentication header must be '{self.header_key}: {self.header_prefix} <token>'"
            )

        if header_prefix != self.header_prefix:
            logger.debug("Wrong prefix '%s' for authorization header, expected '%s'", header_prefix, self.header_prefix)
            raise exceptions.JWTException(
                f"Authentication header must be '{self.header_key}: {self.header_prefix} <token>'"
            )

        return token.encode()

    def _resolve_token(self, headers: Headers, cookies: Cookies) -> jwt.JWT:
        try:
            try:
                encoded_token = self._token_from_header(headers)
            except exceptions.Unauthorized:
                encoded_token = self._token_from_cookies(cookies)
        except exceptions.Unauthorized:
            raise HTTPException(status_code=http.HTTPStatus.UNAUTHORIZED)
        except exceptions.JWTException as e:
            raise HTTPException(
                status_code=http.HTTPStatus.BAD_REQUEST, detail={"error": e.__class__, "description": str(e)}
            )

        try:
            token = jwt.JWT.decode(encoded_token, self.secret)
        except (exceptions.JWTDecodeException, exceptions.JWTValidateException) as e:
            raise HTTPException(
                status_code=http.HTTPStatus.UNAUTHORIZED, detail={"error": e.__class__, "description": str(e)}
            )

        return token


class AccessTokenComponent(BaseTokenComponent):
    def __init__(
        self,
        secret: bytes,
        *,
        header_prefix: str = "Bearer",
        header_key: str = "access_token",
        cookie_key: str = "access_token",
    ):
        super().__init__(secret, header_prefix=header_prefix, header_key=header_key, cookie_key=cookie_key)

    def resolve(self, headers: Headers, cookies: Cookies) -> types.AccessToken:
        return t.cast(types.AccessToken, self._resolve_token(headers, cookies))


class RefreshTokenComponent(BaseTokenComponent):
    def __init__(
        self,
        secret: bytes,
        *,
        header_prefix: str = "Bearer",
        header_key: str = "refresh_token",
        cookie_key: str = "refresh_token",
    ):
        super().__init__(secret, header_prefix=header_prefix, header_key=header_key, cookie_key=cookie_key)

    def resolve(self, headers: Headers, cookies: Cookies) -> types.RefreshToken:
        return t.cast(types.RefreshToken, self._resolve_token(headers, cookies))



import http
import typing as t

import starlette.exceptions

from flama import compat

__all__ = [
    "ApplicationError",
    "DependencyNotInstalled",
    "SQLAlchemyError",
    "DecodeError",
    "HTTPException",
    "NoCodecAvailable",
    "SerializationError",
    "ValidationError",
    "WebSocketException",
    "NotFoundException",
    "MethodNotAllowedException",
    "FrameworkNotInstalled",
    "FrameworkVersionWarning",
]


class ApplicationError(Exception): ...


class DependencyNotInstalled(ApplicationError):
    class Dependency(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
        pydantic = "pydantic"
        marshmallow = "marshmallow"
        apispec = "apispec"
        typesystem = "typesystem"
        sqlalchemy = "sqlalchemy[asyncio]"
        httpx = "httpx"
        tomli = "tomli"

    def __init__(
        self,
        *,
        dependency: t.Optional[t.Union[str, Dependency]] = None,
        dependant: t.Optional[str] = None,
        msg: str = "",
    ) -> None:
        super().__init__()
        self.dependency = self.Dependency(dependency) if dependency else None
        self.dependant = dependant
        self.msg = msg

    def __str__(self) -> str:
        if self.dependency:
            s = f"Dependency '{self.dependency.value}' must be installed"
            if self.dependant:
                s += f" to use '{self.dependant}'"
            if self.msg:
                s += f" ({self.msg})"
        else:
            s = self.msg

        return s

    def __repr__(self) -> str:
        params = ("msg", "dependency", "dependant")
        formatted_params = ", ".join([f"{x}={getattr(self, x)}" for x in params if getattr(self, x)])
        return f"{self.__class__.__name__}({formatted_params})"


class SQLAlchemyError(ApplicationError): ...


class DecodeError(Exception):
    """
    Raised by a Codec when `decode` fails due to malformed syntax.
    """

    def __init__(self, message, marker=None, base_format=None) -> None:
        super().__init__(self, message)
        self.message = message
        self.marker = marker
        self.base_format = base_format


class NoCodecAvailable(Exception): ...


class WebSocketException(starlette.exceptions.WebSocketException):
    def __init__(self, code: int, reason: t.Optional[str] = None) -> None:
        self.code = code
        self.reason = reason or ""

    def __str__(self) -> str:
        return str(self.reason)

    def __repr__(self) -> str:
        params = ("code", "reason")
        formatted_params = ", ".join([f"{x}={getattr(self, x)}" for x in params if getattr(self, x)])
        return f"{self.__class__.__name__}({formatted_params})"

    def __eq__(self, other):
        return isinstance(other, WebSocketException) and self.code == other.code and self.reason == other.reason


class HTTPException(starlette.exceptions.HTTPException):
    def __init__(
        self,
        status_code: int,
        detail: t.Optional[t.Union[str, dict[str, t.Any]]] = None,
        headers: t.Optional[dict] = None,
    ) -> None:
        if detail is None:
            detail = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.detail = detail  # type: ignore[assignment]
        self.headers = headers

    def __str__(self) -> str:
        return str(self.detail)

    def __repr__(self) -> str:
        params = ("status_code", "detail", "headers")
        formatted_params = ", ".join([f"{x}={getattr(self, x)}" for x in params if getattr(self, x)])
        return f"{self.__class__.__name__}({formatted_params})"

    def __eq__(self, other):
        return (
            isinstance(other, HTTPException)
            and self.status_code == other.status_code
            and self.detail == other.detail
            and self.headers == other.headers
        )


class ValidationError(HTTPException):
    def __init__(
        self,
        detail: t.Optional[t.Union[str, dict[str, list[str]]]] = None,
        status_code: int = 400,
    ) -> None:
        super().__init__(status_code, detail=detail)


class SerializationError(HTTPException):
    def __init__(self, detail: t.Union[None, str, dict[str, list[str]]] = None, status_code: int = 500) -> None:
        super().__init__(status_code, detail=detail)


class NotFoundException(Exception):
    def __init__(
        self, path: t.Optional[str] = None, params: t.Optional[dict[str, t.Any]] = None, name: t.Optional[str] = None
    ) -> None:
        self.path = path
        self.params = params
        self.name = name

    def __str__(self) -> str:
        params = ("path", "params", "name")
        formatted_params = ", ".join([f"{x}={repr(getattr(self, x))}" for x in params if getattr(self, x)])
        return f"Path not found ({formatted_params})"

    def __repr__(self) -> str:
        params = ("path", "params", "name")
        formatted_params = ", ".join([f"{x}={repr(getattr(self, x))}" for x in params if getattr(self, x)])
        return f"{self.__class__.__name__}({formatted_params})"


class MethodNotAllowedException(Exception):
    def __init__(self, path: str, method: str, allowed: set[str]) -> None:
        self.path = path
        self.method = method
        self.allowed = allowed

    def __str__(self) -> str:
        params = ("path", "method", "allowed")
        formatted_params = ", ".join([f"{x}={getattr(self, x)}" for x in params if getattr(self, x)])
        return f"Method not allowed ({formatted_params})"

    def __repr__(self) -> str:
        params = ("path", "method", "allowed")
        formatted_params = ", ".join([f"{x}={getattr(self, x)}" for x in params if getattr(self, x)])
        return f"{self.__class__.__name__}({formatted_params})"


class FrameworkNotInstalled(Exception):
    """Cannot find an installed version of the framework."""

    ...


class FrameworkVersionWarning(Warning):
    """Warning for when a framework version does not match."""

    ...



from unittest.mock import MagicMock, call, patch

import pytest

from flama.ddd.repositories import HTTPRepository
from flama.ddd.workers import HTTPWorker


class TestCaseHTTPWorker:
    @pytest.fixture(scope="function")
    def worker(self, client):
        class FooWorker(HTTPWorker):
            bar: HTTPRepository

        return FooWorker("foo", client.app)

    def test_init(self, app):
        worker = HTTPWorker("foo", app)

        assert worker._app == app
        assert worker._url == "foo"
        assert not hasattr(worker, "client")

    def test_client(self, worker):
        with pytest.raises(AttributeError, match="Client not initialized"):
            worker.client

    async def test_set_up(self, worker):
        with patch("flama.client.Client"):
            await worker.set_up()
            assert worker.client.__aenter__.await_args_list == [call()]

    async def test_tear_down(self, worker):
        client_mock = MagicMock()
        worker.client = client_mock

        await worker.tear_down()
        assert client_mock.__aexit__.await_args_list == [call()]

    async def test_begin(self, worker):
        worker.client = MagicMock()

        with patch.object(worker, "set_up"):
            assert not hasattr(worker, "bar")

            await worker.begin()

            assert worker.set_up.await_args_list == [call()]
            assert hasattr(worker, "bar")
            assert isinstance(worker.bar, HTTPRepository)

    @pytest.mark.parametrize(
        ["rollback"],
        (
            pytest.param(True, id="ok_rollback"),
            pytest.param(False, id="ok_commit"),
        ),
    )
    async def test_end(self, worker, rollback):
        worker.bar = MagicMock()
        worker.client = MagicMock()

        with patch.object(worker, "tear_down"):
            assert hasattr(worker, "bar")

            await worker.end(rollback=rollback)

            assert worker.tear_down.await_args_list == [call(rollback=rollback)]
            assert not hasattr(worker, "bar")



import json
import tempfile

import pytest

from flama.config import exceptions
from flama.config.data_structures import FileDict


class TestCaseFileDict:
    @pytest.fixture(scope="function")
    def file_dict(self):
        content = {"foo": "bar"}

        with tempfile.NamedTemporaryFile("w+") as f:
            json.dump(content, f)
            f.flush()
            yield FileDict(f.name, "json")

    @pytest.mark.parametrize(
        ["config_file", "format", "exception"],
        (
            pytest.param("config_no_sections", "ini", None, id="config_no_sections"),
            pytest.param("config", "ini", None, id="config"),
            pytest.param("json", "json", None, id="json"),
            pytest.param("yaml", "yaml", None, id="yaml"),
            pytest.param("toml", "toml", None, id="toml"),
            pytest.param("toml", "wrong", exceptions.ConfigError("Wrong config file format"), id="wrong_format"),
            pytest.param("toml", "json", exceptions.ConfigError("Config file cannot be loaded"), id="cannot_load"),
        ),
        indirect=["config_file", "exception"],
    )
    def test_init(self, config_file, format, exception):
        file_path, config_obj = config_file

        with exception:
            file_dict = FileDict(file_path, format)

            assert file_dict == config_obj

    def test_getitem(self, file_dict):
        assert file_dict["foo"] == "bar"

    def test_eq(self, file_dict):
        assert file_dict == {"foo": "bar"}

    def test_iter(self, file_dict):
        assert list(iter(file_dict)) == ["foo"]

    def test_len(self, file_dict):
        assert len(file_dict) == 1

    def test_repr(self, file_dict):
        assert repr(file_dict) == "FileDict({'foo': 'bar'})"



import abc
import time
import typing as t

from flama.authentication import exceptions

if t.TYPE_CHECKING:
    from flama.authentication.jwt.jwt import Payload

__all__ = [
    "ClaimValidator",
    "IssValidator",
    "SubValidator",
    "AudValidator",
    "ExpValidator",
    "NbfValidator",
    "IatValidator",
    "JtiValidator",
]


class ClaimValidator(abc.ABC):
    claim: t.ClassVar[str]

    def __init__(self, payload: "Payload", claims: dict[str, t.Any]) -> None:
        self.value = claims.get(self.claim)
        self.payload = payload

    @abc.abstractmethod
    def validate(self):
        """Validate the claim.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        ...


class IssValidator(ClaimValidator):
    """Issuer claim validator."""

    claim = "iss"

    def validate(self):
        """Validate the claim.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        ...


class SubValidator(ClaimValidator):
    """Subject claim validator."""

    claim = "sub"

    def validate(self):
        """Validate the claim.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        ...


class AudValidator(ClaimValidator):
    """Audience claim validator."""

    claim = "aud"

    def validate(self):
        """Validate the claim.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        ...


class ExpValidator(ClaimValidator):
    """Expiration time claim validator.

    The value of the claim must be a number representing the expiration time of the token in seconds since the epoch
    (UTC). The expiration time must be after the current time.
    """

    claim = "exp"

    def validate(self):
        """Validate the claim.

        The value of the claim must be a number representing the expiration time of the token in seconds since the
        epoch (UTC). The expiration time must be after the current time.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        if self.payload.exp is not None and self.payload.exp < int(time.time()):
            raise exceptions.JWTClaimValidateException("exp")


class NbfValidator(ClaimValidator):
    """Not before claim validator.

    The value of the claim must be a number representing the time before which the token must not be accepted for
    processing in seconds since the epoch (UTC). The time must be before the current time.
    """

    claim = "nbf"

    def validate(self):
        """Validate the claim.

        The value of the claim must be a number representing the time before which the token must not be accepted for
        processing in seconds since the epoch (UTC). The time must be before the current time.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        if self.payload.nbf is not None and self.payload.nbf > int(time.time()):
            raise exceptions.JWTClaimValidateException("nbf")


class IatValidator(ClaimValidator):
    """Issued at claim validator.

    The value of the claim must be a number representing the time at which the JWT was issued in seconds since the
    epoch (UTC). The time must be before the current time.
    """

    claim = "iat"

    def validate(self):
        """Validate the claim.

        The value of the claim must be a number representing the time at which the JWT was issued in seconds since the
        epoch (UTC). The time must be before the current time.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        if self.payload.iat is not None and self.payload.iat > int(time.time()):
            raise exceptions.JWTClaimValidateException("iat")


class JtiValidator(ClaimValidator):
    """JWT ID claim validator."""

    claim = "jti"

    def validate(self):
        """Validate the claim.

        :raises JWTClaimValidateException: if the claim is not valid.
        """
        ...



import dataclasses
import typing as t

try:
    from sqlalchemy import Table
except Exception:  # pragma: no cover
    Table = t.Any

__all__ = ["Model", "PrimaryKey", "Schema", "Metadata", "MethodMetadata"]


@dataclasses.dataclass
class PrimaryKey:
    name: str
    type: type


@dataclasses.dataclass
class Model:
    table: Table  # type: ignore
    primary_key: PrimaryKey


@dataclasses.dataclass
class Schema:
    name: str
    schema: t.Any


@dataclasses.dataclass
class Schemas:
    input: Schema
    output: Schema


@dataclasses.dataclass
class Metadata:
    name: str = dataclasses.field(init=False)
    verbose_name: str = dataclasses.field(init=False)
    namespaces: dict[str, dict[str, t.Any]] = dataclasses.field(default_factory=dict)

    def to_plain_dict(self) -> dict[str, t.Any]:
        return {
            "name": self.name,
            "verbose_name": self.verbose_name,
            **{f"{namespace}_{k}": v for namespace, values in self.namespaces.items() for k, v in values.items()},
        }


@dataclasses.dataclass
class MethodMetadata:
    path: str
    methods: set[str] = dataclasses.field(default_factory=lambda: {"GET"})
    name: t.Optional[str] = None
    tags: dict[str, t.Any] = dataclasses.field(default_factory=dict)



import typing as t

from flama import concurrency, http, types
from flama.endpoints.base import BaseEndpoint

__all__ = ["HTTPEndpoint"]


class HTTPEndpoint(BaseEndpoint, types.HTTPEndpointProtocol):
    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        """An HTTP endpoint.

        :param scope: ASGI scope.
        :param receive: ASGI receive function.
        :param send: ASGI send function.
        """
        if scope["type"] != "http":
            raise ValueError("Wrong scope")

        super().__init__(scope, receive, send)

        self.state.update(
            {
                "request": http.Request(self.state["scope"], receive=receive),
            }
        )

    @classmethod
    def allowed_methods(cls) -> set[str]:
        """The list of allowed methods by this endpoint.

        :return: List of allowed methods.
        """
        methods = {
            method for method in http.Method.__members__.keys() if getattr(cls, method.lower(), None) is not None
        }
        if "GET" in methods:
            methods.add("HEAD")
        return methods

    @classmethod
    def allowed_handlers(cls) -> dict[str, t.Callable]:
        """A mapping of handler related to each HTTP method.

        :return: Handlers mapping.
        """
        return {method: getattr(cls, method.lower(), getattr(cls, "get")) for method in cls.allowed_methods()}

    @property
    def handler(self) -> t.Callable:
        """The handler used for dispatching this request.

        :return: Handler.
        """
        handler_name = "get" if self.state["request"].method == "HEAD" else self.state["request"].method.lower()
        h: t.Callable = getattr(self, handler_name)
        return h

    async def dispatch(self) -> None:
        """Dispatch a request."""
        app = self.state["app"]
        handler = await app.injector.inject(self.handler, self.state)
        return await concurrency.run(handler)



import inspect
import logging
import typing as t

from flama import compat, concurrency, endpoints, exceptions, http, schemas, types
from flama.routing.routes.base import BaseEndpointWrapper, BaseRoute

__all__ = ["Route"]

logger = logging.getLogger(__name__)


class BaseHTTPEndpointWrapper(BaseEndpointWrapper):
    def __init__(self, handler: types.Handler, *, pagination: t.Optional[types.Pagination] = None):
        super().__init__(handler, pagination=pagination)

        try:
            self.schema = schemas.Schema.from_type(inspect.signature(self.handler).return_annotation).unique_schema
        except Exception:
            self.schema = None

    def _build_api_response(self, response: t.Union[http.Response, None]) -> http.Response:
        """Build an API response given a handler and the current response.

        It infers the output schema from the handler signature or just wraps the response in a APIResponse object.

        :param response: The current response.
        :return: An API response.
        """
        if isinstance(response, (dict, list)):
            response = http.APIResponse(content=response, schema=self.schema)
        elif isinstance(response, (str, bytes)):
            response = http.APIResponse(content=response)
        elif response is None:
            response = http.APIResponse(content="")

        return response


class HTTPFunctionWrapper(BaseHTTPEndpointWrapper):
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        """
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "route": route,
            "request": http.Request(route_scope, receive=receive),
        }

        injected_func = await app.injector.inject(self.handler, state)
        response = await concurrency.run(injected_func)
        response = self._build_api_response(response)

        await response(route_scope, receive, send)


class HTTPEndpointWrapper(BaseHTTPEndpointWrapper):
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        """
        response = await self.handler(scope, receive, send)
        response = self._build_api_response(response)

        await response(scope, receive, send)


class Route(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: types.HTTPHandler,
        *,
        methods: t.Optional[t.Union[set[str], t.Sequence[str]]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        pagination: t.Optional[types.Pagination] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> None:
        """A route definition of a http endpoint.

        :param path: URL path.
        :param endpoint: HTTP endpoint or function.
        :param methods: List of valid HTTP methods.
        :param name: Route name.
        :param include_in_schema: True if this route must be listed as part of the App schema.
        :param pagination: Apply a pagination technique.
        :param tags: Route tags.
        """
        if not (self.is_endpoint(endpoint) or (not inspect.isclass(endpoint) and callable(endpoint))):
            raise exceptions.ApplicationError("Endpoint must be a callable or an HTTPEndpoint subclass")

        if self.is_endpoint(endpoint):
            self.methods = endpoint.allowed_methods() if methods is None else set(methods)
        else:
            self.methods = {"GET"} if methods is None else set(methods)

        if "GET" in self.methods:
            self.methods.add("HEAD")

        name = endpoint.__name__ if name is None else name
        wrapper = HTTPEndpointWrapper if inspect.isclass(endpoint) else HTTPFunctionWrapper

        super().__init__(
            path, wrapper(endpoint, pagination=pagination), name=name, include_in_schema=include_in_schema, tags=tags
        )

        self.app: BaseEndpointWrapper

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if scope["type"] == "http":
            await self.handle(types.Scope({**scope, **self.route_scope(scope)}), receive, send)

    def __eq__(self, other: t.Any) -> bool:
        return super().__eq__(other) and isinstance(other, Route) and self.methods == other.methods

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r}, name={self.name!r}, methods={sorted(self.methods)!r})"

    @staticmethod
    def is_endpoint(
        x: t.Union[t.Callable, type[endpoints.HTTPEndpoint]],
    ) -> compat.TypeGuard[type[endpoints.HTTPEndpoint]]:  # PORT: Replace compat when stop supporting 3.9
        return inspect.isclass(x) and issubclass(x, endpoints.HTTPEndpoint)

    def endpoint_handlers(self) -> dict[str, t.Callable]:
        """Return a mapping of all possible endpoints of this route.

        Useful to identify all endpoints by HTTP methods.

        :return: Mapping of all endpoints.
        """
        if self.is_endpoint(self.endpoint):
            return {
                method: handler
                for method, handler in self.endpoint.allowed_handlers().items()
                if method in self.methods
            }

        return {method: self.endpoint for method in self.methods}

    def match(self, scope: types.Scope) -> BaseRoute.Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        if scope["type"] != "http":
            return self.Match.none

        m = super().match(scope)
        if m == self.Match.none:
            return m

        return self.Match.full if scope["method"] in self.methods else self.Match.partial



import typing

import tensorflow as tf

import flama
from flama import Flama
from flama.models import ModelComponent, ModelResource, ModelResourceType
from flama.models.base import Model
from flama.resources.routing import resource_method

app = Flama()

# Adding a model:
app.models.add_model(
    "/model",
    "path/to/your_model_file.flm",
    "model",
)


# Adding a model using a ModelResource:
@app.models.model("/model_resource")
class TensorFlowModelResource(ModelResource, metaclass=ModelResourceType):
    name = "tensorflow_model"
    verbose_name = "TensorFlow Logistic Regression"
    model_path = "path/to/your_model_file.flm"

    @resource_method("/info", methods=["GET"], name="model-info")
    def info(self):
        return {"name": self.verbose_name}


# Adding a model using a custom component:
model = tf.keras.models.Sequential(
    [
        tf.keras.layers.Flatten(input_shape=(28, 28)),
        tf.keras.layers.Dense(128, activation="relu"),
        tf.keras.layers.Dropout(0.2),
        tf.keras.layers.Dense(10, activation="softmax"),
    ]
)


class CustomModel(Model):
    def inspect(self) -> typing.Any:
        return self.model.__dict__

    def predict(self, x: typing.Any) -> typing.Any:
        return self.model.predict(x)


class CustomModelComponent(ModelComponent):
    def resolve(self) -> CustomModel:
        return self.model


component = CustomModelComponent(model)


class CustomModelResource(ModelResource, metaclass=ModelResourceType):
    name = "custom_model"
    verbose_name = "Custom model"
    component = component


app.add_component(component)
app.models.add_model_resource("/", CustomModelResource)

if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



import contextlib
import typing as t
from collections import namedtuple

import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields

from flama import schemas
from flama.endpoints import HTTPEndpoint
from flama.routing import Router
from flama.schemas import openapi
from flama.schemas.generator import SchemaRegistry
from tests.asserts import assert_recursive_contains


def _fix_ref(value, refs):
    try:
        prefix, name = value.rsplit("/", 1)
        return f"{prefix}/{refs[name]}"
    except KeyError:
        return value


def _replace_refs(schema, refs):
    if isinstance(schema, dict):
        return {k: _fix_ref(v, refs) if k == "$ref" else _replace_refs(v, refs) for k, v in schema.items()}

    if isinstance(schema, (list, tuple, set)):
        return [_replace_refs(x, refs) for x in schema]

    return schema


class TestCaseSchemaRegistry:
    @pytest.fixture(scope="function")
    def registry(self, app):
        return SchemaRegistry()

    @pytest.fixture(scope="function")
    def spec(self, openapi_spec):
        return openapi.OpenAPISpec.from_spec(openapi_spec)

    def test_empty_init(self):
        assert SchemaRegistry() == {}

    @pytest.mark.parametrize(
        ["operation", "register_schemas", "output"],
        [
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Foo",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Reference(ref="#!/components/schemas/Foo")
                                    )
                                },
                            )
                        }
                    )
                ),
                ["Foo"],
                ["Foo"],
                id="response_reference",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Bar",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Reference(ref="#!/components/schemas/Bar")
                                    )
                                },
                            )
                        }
                    )
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="response_reference_nested",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Bar",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Reference(ref="#!/components/schemas/BarList")
                                    )
                                },
                            )
                        }
                    )
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="response_reference_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Bar",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Reference(ref="#!/components/schemas/BarDict")
                                    )
                                },
                            )
                        }
                    )
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="response_reference_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Foo",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Schema(
                                            {
                                                "type": "object",
                                                "properties": {"foo": {"$ref": "#!/components/schemas/Foo"}},
                                            }
                                        ),
                                    )
                                },
                            )
                        }
                    )
                ),
                ["Foo"],
                ["Foo"],
                id="response_schema",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Bar",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Schema(
                                            {
                                                "type": "object",
                                                "properties": {"bar": {"$ref": "#!/components/schemas/Bar"}},
                                            }
                                        ),
                                    )
                                },
                            )
                        }
                    )
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="response_schema_nested",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="BarList",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Schema(
                                            {
                                                "type": "object",
                                                "properties": {"bar": {"$ref": "#!/components/schemas/BarList"}},
                                            }
                                        ),
                                    )
                                },
                            )
                        }
                    )
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="response_schema_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="BarDict",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Schema(
                                            {
                                                "type": "object",
                                                "properties": {"bar": {"$ref": "#!/components/schemas/BarDict"}},
                                            }
                                        ),
                                    )
                                },
                            )
                        }
                    )
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="response_schema_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Foo",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Schema(
                                            {
                                                "type": "object",
                                                "properties": {
                                                    "foo": {
                                                        "type": "array",
                                                        "items": {"$ref": "#!/components/schemas/Foo"},
                                                    }
                                                },
                                            }
                                        ),
                                    )
                                },
                            )
                        }
                    )
                ),
                ["Foo"],
                ["Foo"],
                id="response_array",
            ),
            pytest.param(
                openapi.Operation(
                    responses=openapi.Responses(
                        {
                            "200": openapi.Response(
                                description="Foo",
                                content={
                                    "application/json": openapi.MediaType(
                                        schema=openapi.Schema(
                                            {"type": "object", "properties": {"foo": {"type": "foo"}}}
                                        ),
                                    )
                                },
                            )
                        }
                    )
                ),
                [],
                [],
                id="response_wrong",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.Reference(ref="#!/components/schemas/Foo"), responses=openapi.Responses({})
                ),
                ["Foo"],
                ["Foo"],
                id="body_reference",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.Reference(ref="#!/components/schemas/Bar"), responses=openapi.Responses({})
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="body_reference_nested",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.Reference(ref="#!/components/schemas/BarList"), responses=openapi.Responses({})
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="body_reference_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.Reference(ref="#!/components/schemas/BarDict"), responses=openapi.Responses({})
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="body_reference_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.RequestBody(
                        description="Foo",
                        content={
                            "application/json": openapi.MediaType(
                                schema=openapi.Schema(
                                    {"type": "object", "properties": {"foo": {"$ref": "#!/components/schemas/Foo"}}}
                                ),
                            )
                        },
                    ),
                    responses=openapi.Responses({}),
                ),
                ["Foo"],
                ["Foo"],
                id="body_schema",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.RequestBody(
                        description="Bar",
                        content={
                            "application/json": openapi.MediaType(
                                schema=openapi.Schema(
                                    {"type": "object", "properties": {"foo": {"$ref": "#!/components/schemas/Bar"}}}
                                ),
                            )
                        },
                    ),
                    responses=openapi.Responses({}),
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="body_schema_nested",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.RequestBody(
                        description="BarList",
                        content={
                            "application/json": openapi.MediaType(
                                schema=openapi.Schema(
                                    {"type": "object", "properties": {"foo": {"$ref": "#!/components/schemas/BarList"}}}
                                ),
                            )
                        },
                    ),
                    responses=openapi.Responses({}),
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="body_schema_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.RequestBody(
                        description="BarDict",
                        content={
                            "application/json": openapi.MediaType(
                                schema=openapi.Schema(
                                    {"type": "object", "properties": {"foo": {"$ref": "#!/components/schemas/BarDict"}}}
                                ),
                            )
                        },
                    ),
                    responses=openapi.Responses({}),
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="body_schema_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.RequestBody(
                        description="Foo",
                        content={
                            "application/json": openapi.MediaType(
                                schema=openapi.Schema(
                                    {
                                        "type": "object",
                                        "properties": {
                                            "foo": {"type": "array", "items": {"$ref": "#!/components/schemas/Foo"}}
                                        },
                                    }
                                ),
                            )
                        },
                    ),
                    responses=openapi.Responses({}),
                ),
                ["Foo"],
                ["Foo"],
                id="body_array",
            ),
            pytest.param(
                openapi.Operation(
                    requestBody=openapi.RequestBody(
                        description="Foo",
                        content={
                            "application/json": openapi.MediaType(
                                schema=openapi.Schema({"type": "object", "properties": {"foo": {"type": "foo"}}}),
                            )
                        },
                    ),
                    responses=openapi.Responses({}),
                ),
                [],
                [],
                id="body_wrong",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[openapi.Reference(ref="#!/components/schemas/Foo")], responses=openapi.Responses({})
                ),
                ["Foo"],
                ["Foo"],
                id="parameter_reference",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[openapi.Reference(ref="#!/components/schemas/Bar")], responses=openapi.Responses({})
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="parameter_reference_nested",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[openapi.Reference(ref="#!/components/schemas/BarList")], responses=openapi.Responses({})
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="parameter_reference_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[openapi.Reference(ref="#!/components/schemas/BarDict")], responses=openapi.Responses({})
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="parameter_reference_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[
                        openapi.Parameter(
                            in_="query",
                            name="foo",
                            schema=openapi.Schema(
                                {"type": "object", "properties": {"foo": {"$ref": "#!/components/schemas/Foo"}}}
                            ),
                        )
                    ],
                    responses=openapi.Responses({}),
                ),
                ["Foo"],
                ["Foo"],
                id="parameter_schema",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[
                        openapi.Parameter(
                            in_="query",
                            name="bar",
                            schema=openapi.Schema(
                                {"type": "object", "properties": {"bar": {"$ref": "#!/components/schemas/Bar"}}}
                            ),
                        )
                    ],
                    responses=openapi.Responses({}),
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="parameter_schema_nested",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[
                        openapi.Parameter(
                            in_="query",
                            name="bar",
                            schema=openapi.Schema(
                                {"type": "object", "properties": {"bar": {"$ref": "#!/components/schemas/BarList"}}}
                            ),
                        )
                    ],
                    responses=openapi.Responses({}),
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="parameter_schema_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[
                        openapi.Parameter(
                            in_="query",
                            name="bar",
                            schema=openapi.Schema(
                                {"type": "object", "properties": {"bar": {"$ref": "#!/components/schemas/BarDict"}}}
                            ),
                        )
                    ],
                    responses=openapi.Responses({}),
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="parameter_schema_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[
                        openapi.Parameter(
                            in_="query",
                            name="foo",
                            schema=openapi.Schema(
                                {
                                    "type": "object",
                                    "properties": {
                                        "foo": {"type": "array", "items": {"$ref": "#!/components/schemas/Foo"}}
                                    },
                                }
                            ),
                        )
                    ],
                    responses=openapi.Responses({}),
                ),
                ["Foo"],
                ["Foo"],
                id="parameter_array",
            ),
            pytest.param(
                openapi.Operation(
                    parameters=[
                        openapi.Parameter(
                            in_="query",
                            name="foo",
                            schema=openapi.Schema({"type": "object", "properties": {"foo": {"type": "foo"}}}),
                        )
                    ],
                    responses=openapi.Responses({}),
                ),
                [],
                [],
                id="parameter_wrong",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={"200": openapi.Reference(ref="#!/components/schemas/Foo")},
                    responses=openapi.Responses({}),
                ),
                ["Foo"],
                ["Foo"],
                id="callback_reference",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={"200": openapi.Reference(ref="#!/components/schemas/Bar")},
                    responses=openapi.Responses({}),
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="callback_reference_nested",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={"200": openapi.Reference(ref="#!/components/schemas/BarList")},
                    responses=openapi.Responses({}),
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="callback_reference_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={"200": openapi.Reference(ref="#!/components/schemas/BarDict")},
                    responses=openapi.Responses({}),
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="callback_reference_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={
                        "foo": openapi.Callback(
                            {
                                "/callback": openapi.Path(
                                    get=openapi.Operation(
                                        responses=openapi.Responses(
                                            {
                                                "200": openapi.Response(
                                                    description="Foo",
                                                    content={
                                                        "application/json": openapi.MediaType(
                                                            schema=openapi.Schema(
                                                                {
                                                                    "type": "object",
                                                                    "properties": {
                                                                        "foo": {"$ref": "#!/components/schemas/Foo"}
                                                                    },
                                                                }
                                                            )
                                                        )
                                                    },
                                                )
                                            }
                                        )
                                    )
                                )
                            }
                        )
                    },
                    responses=openapi.Responses({}),
                ),
                ["Foo"],
                ["Foo"],
                id="callback_schema",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={
                        "foo": openapi.Callback(
                            {
                                "/callback": openapi.Path(
                                    get=openapi.Operation(
                                        responses=openapi.Responses(
                                            {
                                                "200": openapi.Response(
                                                    description="Bar",
                                                    content={
                                                        "application/json": openapi.MediaType(
                                                            schema=openapi.Schema(
                                                                {
                                                                    "type": "object",
                                                                    "properties": {
                                                                        "bar": {"$ref": "#!/components/schemas/Bar"}
                                                                    },
                                                                }
                                                            )
                                                        )
                                                    },
                                                )
                                            }
                                        )
                                    )
                                )
                            }
                        )
                    },
                    responses=openapi.Responses({}),
                ),
                ["Bar"],
                ["Foo", "Bar"],
                id="callback_schema_nested",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={
                        "foo": openapi.Callback(
                            {
                                "/callback": openapi.Path(
                                    get=openapi.Operation(
                                        responses=openapi.Responses(
                                            {
                                                "200": openapi.Response(
                                                    description="BarList",
                                                    content={
                                                        "application/json": openapi.MediaType(
                                                            schema=openapi.Schema(
                                                                {
                                                                    "type": "object",
                                                                    "properties": {
                                                                        "bar": {"$ref": "#!/components/schemas/BarList"}
                                                                    },
                                                                }
                                                            )
                                                        )
                                                    },
                                                )
                                            }
                                        )
                                    )
                                )
                            }
                        )
                    },
                    responses=openapi.Responses({}),
                ),
                ["BarList"],
                ["Foo", "BarList"],
                id="callback_schema_nested_list",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={
                        "foo": openapi.Callback(
                            {
                                "/callback": openapi.Path(
                                    get=openapi.Operation(
                                        responses=openapi.Responses(
                                            {
                                                "200": openapi.Response(
                                                    description="BarDict",
                                                    content={
                                                        "application/json": openapi.MediaType(
                                                            schema=openapi.Schema(
                                                                {
                                                                    "type": "object",
                                                                    "properties": {
                                                                        "bar": {"$ref": "#!/components/schemas/BarDict"}
                                                                    },
                                                                }
                                                            )
                                                        )
                                                    },
                                                )
                                            }
                                        )
                                    )
                                )
                            }
                        )
                    },
                    responses=openapi.Responses({}),
                ),
                ["BarDict"],
                ["Foo", "BarDict"],
                id="callback_schema_nested_dict",
            ),
            pytest.param(
                openapi.Operation(
                    callbacks={
                        "foo": openapi.Callback(
                            {
                                "/callback": openapi.Path(
                                    get=openapi.Operation(
                                        responses=openapi.Responses(
                                            {
                                                "200": openapi.Response(
                                                    description="Foo",
                                                    content={
                                                        "application/json": openapi.MediaType(
                                                            schema=openapi.Schema(
                                                                {
                                                                    "type": "object",
                                                                    "properties": {"foo": {"type": "foo"}},
                                                                }
                                                            )
                                                        )
                                                    },
                                                )
                                            }
                                        )
                                    )
                                )
                            }
                        )
                    },
                    responses=openapi.Responses({}),
                ),
                [],
                [],
                id="callback_wrong",
            ),
        ],
    )
    def test_used(self, registry, schemas, spec, operation, register_schemas, output):
        for schema in register_schemas:
            registry.register(schemas[schema].schema, name=schema)

        expected_output = {id(schemas[schema].schema) for schema in output}

        spec.add_path("/", openapi.Path(get=operation))

        assert set(registry.used(spec).keys()) == expected_output

    @pytest.mark.parametrize(
        ["schema", "explicit_name", "output"],
        [
            pytest.param("Foo", "Foo", {"Foo": "Foo"}, id="explicit_name"),
            pytest.param("Foo", None, {"Foo": None}, id="infer_name"),
            pytest.param("Bar", "Bar", {"Bar": "Bar"}, id="nested_schemas"),
        ],
    )
    def test_register(self, registry, schemas, schema, explicit_name, output):
        schema, name = schemas[schema]
        expected_name = name if not explicit_name else explicit_name
        exception = (
            contextlib.ExitStack() if expected_name else pytest.raises(ValueError, match="Cannot infer schema name.")
        )
        with exception:
            registry.register(schema, name=explicit_name)
            for s, n in output.items():
                assert schemas[s].schema in registry
                assert registry[schemas[s].schema].name == (n or schemas[s].name)

    def test_register_already_registered(self, registry, foo_schema):
        schema = foo_schema.schema
        registry.register(schema, name="Foo")
        with pytest.raises(ValueError, match="Schema is already registered."):
            registry.register(schema, name="Foo")

    @pytest.mark.parametrize(
        ["multiple", "result"],
        (
            pytest.param(
                False,
                openapi.Reference(ref="#/components/schemas/Foo"),
                id="schema",
            ),
            pytest.param(
                True,
                openapi.Schema({"type": "array", "items": {"$ref": "#/components/schemas/Foo"}}),
                id="array",
            ),
        ),
    )
    def test_get_openapi_ref(self, multiple, result, registry, foo_schema):
        schema = foo_schema.schema
        registry.register(schema, name="Foo")
        assert registry.get_openapi_ref(schema, multiple=multiple) == result


class TestCaseSchemaGenerator:
    @pytest.fixture(scope="function")
    def owner_schema(self, app):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("Owner", name=(str, ...), __module__="pydantic.main")
            name = "pydantic.main.Owner"
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(title="Owner", fields={"name": typesystem.fields.String()})
            name = "typesystem.schemas.Owner"
        elif app.schema.schema_library.lib == marshmallow:
            schema = type("Owner", (marshmallow.Schema,), {"name": marshmallow.fields.String()})
            name = "abc.Owner"
        else:
            raise ValueError("Wrong schema lib")
        return namedtuple("OwnerSchema", ("schema", "name"))(schema, name)

    @pytest.fixture(scope="function")
    def puppy_schema(self, app, owner_schema):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model(
                "Puppy", name=(str, ...), owner=(owner_schema.schema, ...), __module__="pydantic.main"
            )
            name = "pydantic.main.Puppy"
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(
                title="Puppy",
                fields={
                    "name": typesystem.fields.String(),
                    "owner": typesystem.Reference(
                        to="Owner", definitions=typesystem.Definitions({"Owner": owner_schema.schema})
                    ),
                },
            )
            name = "typesystem.schemas.Puppy"
        elif app.schema.schema_library.lib == marshmallow:
            schema = type(
                "Puppy",
                (marshmallow.Schema,),
                {
                    "name": marshmallow.fields.String(),
                    "owner": marshmallow.fields.Nested(owner_schema.schema),
                },
            )
            name = "abc.Puppy"
        else:
            raise ValueError("Wrong schema lib")
        return namedtuple("PuppySchema", ("schema", "name"))(schema, name)

    @pytest.fixture(scope="function")
    def body_param_schema(self, app):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("BodyParam", name=(str, ...), __module__="pydantic.main")
            name = "pydantic.main.BodyParam"
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(title="BodyParam", fields={"name": typesystem.fields.String()})
            name = "typesystem.schemas.BodyParam"
        elif app.schema.schema_library.lib == marshmallow:
            schema = type("BodyParam", (marshmallow.Schema,), {"name": marshmallow.fields.String()})
            name = "abc.BodyParam"
        else:
            raise ValueError("Wrong schema lib")
        return namedtuple("BodyParamSchema", ("schema", "name"))(schema, name)

    @pytest.fixture(scope="function")
    def schemas(self, owner_schema, puppy_schema, body_param_schema):
        return {"Owner": owner_schema, "Puppy": puppy_schema, "BodyParam": body_param_schema}

    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app, puppy_schema, body_param_schema):  # noqa: C901
        @app.route("/endpoint/", methods=["GET"])
        class PuppyEndpoint(HTTPEndpoint):
            async def get(self) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(puppy_schema.schema)]:
                """
                description: Endpoint.
                responses:
                  200:
                    description: Component.
                """
                return {"name": "Canna"}

        @app.route("/custom-component/", methods=["GET"])
        async def get() -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(puppy_schema.schema)]:
            """
            description: Custom component.
            responses:
              200:
                description: Component.
            """
            return {"name": "Canna"}

        @app.route("/many-components/", methods=["GET"])
        async def many_components() -> t.Annotated[
            list[schemas.SchemaType], schemas.SchemaMetadata(puppy_schema.schema)
        ]:
            """
            description: Many custom components.
            responses:
              200:
                description: Many components.
            """
            return [{"name": "foo"}, {"name": "bar"}]

        @app.route("/query-param/", methods=["GET"])
        async def query_param(param1: int, param2: t.Optional[str] = None, param3: bool = True):
            """
            description: Query param.
            responses:
              200:
                description: Param.
            """
            return {"name": param2}

        @app.route("/path-param/{param:int}/", methods=["GET"])
        async def path_param(param: int):
            """
            description: Path param.
            responses:
              200:
                description: Param.
            """
            return {"name": param}

        @app.route("/body-param/", methods=["POST"])
        async def body_param(param: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(body_param_schema.schema)]):
            """
            description: Body param.
            responses:
              200:
                description: Param.
            """
            return {"name": param["name"]}

        @app.route("/default-response/", methods=["GET"])
        async def default_response():
            """
            description: Default response.
            """
            return {"name": "Canna"}

        @app.route("/multiple-responses/", methods=["GET"])
        async def multiple_responses():
            """
            description: Multiple responses.
            responses:
              200:
                description: OK.
              400:
                description: Bad Request.
            """
            return {}

        @app.route("/custom-params/", methods=["GET"])
        async def custom_params(name: str):
            """
            description: Custom params.
            parameters:
              - in: query
                name: name
                schema:
                    type: string
                required: true
                description: The query param
            responses:
              200:
                description: OK.
            """
            return {}

        @app.route("/custom-body/", methods=["POST"])
        async def custom_body():
            """
            description: Custom body.
            requestBody:
                description: The body request
                required: true
                content:
                    multipart/form-data:
                        schema:
                            type: object
                            properties:
                                model:
                                    type: string
                                    format: binary
                                    description: The model file
                                name:
                                    type: string
                                    description: The model name
                                description:
                                    type: string
                                    description: The model description
            responses:
              200:
                description: OK.
            """
            return {}

        @app.route("/wrong-schema/", methods=["POST"])
        async def wrong():
            """
            Wrong schema.
            """
            return {}

        router = Router()
        router.add_route("/custom-component/", endpoint=get, methods=["GET"])
        app.mount("/mount", router)

        nested_router = Router()
        nested_router.add_route("/nested-component/", endpoint=get, methods=["GET"])
        mounted_router = Router()
        mounted_router.mount("/mount", nested_router)
        app.mount("/nested", mounted_router)

    def test_schema_info(self, app):
        schema = app.schema.schema["info"]

        assert schema["title"] == "Foo"
        assert schema["version"] == "1.0.0"
        assert schema["description"] == "Bar"

    def test_components_schemas(self, app, schemas):
        used_schemas = app.schema.schema["components"]["schemas"]

        # Check declared components are only those that are in use
        assert set(used_schemas.keys()) == {
            schemas["Owner"].name,
            schemas["Puppy"].name,
            schemas["BodyParam"].name,
            "flama.APIError",
        }

    @pytest.mark.parametrize(
        "path,verb,expected_schema",
        [
            pytest.param(
                "/query-param/",
                "get",
                {
                    "description": "Query param.",
                    "responses": {
                        "200": {"description": "Param."},
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                    "parameters": [
                        {
                            "name": "param1",
                            "in": "query",
                            "required": True,
                            "schema": {"type": "integer"},
                        },
                        {
                            "name": "param2",
                            "in": "query",
                            "required": False,
                            "schema": {"anyOf": [{"type": "string"}, {"type": "null"}]},
                        },
                        {
                            "name": "param3",
                            "in": "query",
                            "required": False,
                            "schema": {"type": "boolean", "default": True},
                        },
                    ],
                },
                id="query_param",
            ),
            pytest.param(
                "/path-param/{param}/",
                "get",
                {
                    "description": "Path param.",
                    "responses": {
                        "200": {"description": "Param."},
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                    "parameters": [
                        {
                            "name": "param",
                            "in": "path",
                            "required": True,
                            "schema": {"type": "integer"},
                        }
                    ],
                },
                id="path_param",
            ),
            pytest.param(
                "/body-param/",
                "post",
                {
                    "description": "Body param.",
                    "responses": {
                        "200": {"description": "Param."},
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                    "requestBody": {
                        "content": {"application/json": {"schema": {"$ref": "#/components/schemas/BodyParam"}}}
                    },
                },
                id="body_param",
            ),
            pytest.param(
                "/custom-component/",
                "get",
                {
                    "description": "Custom component.",
                    "responses": {
                        "200": {
                            "description": "Component.",
                            "content": {"application/json": {"schema": {"$ref": "#/components/schemas/Puppy"}}},
                        },
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="response_schema_single",
            ),
            pytest.param(
                "/many-components/",
                "get",
                {
                    "description": "Many custom components.",
                    "responses": {
                        "200": {
                            "description": "Many components.",
                            "content": {"application/json": {"schema": {"$ref": "#/components/schemas/Puppy"}}},
                        },
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="response_schema_multiple",
            ),
            pytest.param(
                "/endpoint/",
                "get",
                {
                    "description": "Endpoint.",
                    "responses": {
                        "200": {
                            "description": "Component.",
                            "content": {"application/json": {"schema": {"$ref": "#/components/schemas/Puppy"}}},
                        },
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="response_schema_endpoint",
            ),
            pytest.param(
                "/mount/custom-component/",
                "get",
                {
                    "description": "Custom component.",
                    "responses": {
                        "200": {
                            "description": "Component.",
                            "content": {"application/json": {"schema": {"$ref": "#/components/schemas/Puppy"}}},
                        },
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="response_schema_mount",
            ),
            pytest.param(
                "/nested/mount/nested-component/",
                "get",
                {
                    "description": "Custom component.",
                    "responses": {
                        "200": {
                            "description": "Component.",
                            "content": {"application/json": {"schema": {"$ref": "#/components/schemas/Puppy"}}},
                        },
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="response_schema_nested_mount",
            ),
            pytest.param(
                "/default-response/",
                "get",
                {
                    "description": "Default response.",
                    "responses": {
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        }
                    },
                },
                id="default_response_schema",
            ),
            pytest.param(
                "/multiple-responses/",
                "get",
                {
                    "description": "Multiple responses.",
                    "responses": {
                        "200": {"description": "OK."},
                        "400": {"description": "Bad Request."},
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="multiple_responses",
            ),
            pytest.param(
                "/custom-params/",
                "get",
                {
                    "description": "Custom params.",
                    "parameters": [
                        {
                            "in": "query",
                            "name": "name",
                            "schema": {"type": "string"},
                            "required": True,
                            "description": "The query param",
                        }
                    ],
                    "responses": {
                        "200": {"description": "OK."},
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="custom_params",
            ),
            pytest.param(
                "/custom-body/",
                "post",
                {
                    "description": "Custom body.",
                    "requestBody": {
                        "content": {
                            "multipart/form-data": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "model": {
                                            "type": "string",
                                            "format": "binary",
                                            "description": "The model file",
                                        },
                                        "name": {"type": "string", "description": "The model name"},
                                        "description": {"type": "string", "description": "The model description"},
                                    },
                                }
                            }
                        }
                    },
                    "responses": {
                        "200": {"description": "OK."},
                        "default": {
                            "description": "Unexpected error.",
                            "content": {
                                "application/json": {"schema": {"$ref": "#/components/schemas/flama.APIError"}}
                            },
                        },
                    },
                },
                id="custom_body",
            ),
            pytest.param(
                "/wrong-schema/",
                "post",
                {},
                id="wrong_schema",
            ),
        ],
    )
    def test_schema(self, app, schemas, path, verb, expected_schema):
        schema = _replace_refs(expected_schema, {k: v.name for k, v in schemas.items()})

        assert_recursive_contains(schema, app.schema.schema["paths"][path][verb])



import abc
import dataclasses
import decimal
import enum
import re
import typing as t
import urllib.parse
import uuid

T = t.TypeVar("T", bound=t.Union[int, str, float, decimal.Decimal, uuid.UUID])
FragmentType = t.Literal["constant", "rest", "str", "int", "float", "decimal", "uuid"]

__all__ = ["Path", "URL"]


class Serializer(t.Generic[T], metaclass=abc.ABCMeta):
    regex: t.ClassVar[re.Pattern]
    type: t.ClassVar[type]

    @abc.abstractmethod
    def load(self, value: str) -> T: ...

    @abc.abstractmethod
    def dump(self, value: T) -> str: ...

    def __eq__(self, other):
        return type(other) == type(self)


class StringSerializer(Serializer[str]):
    regex = re.compile(r"[^/]+")
    type = str

    def load(self, value: str) -> str:
        return str(value)

    def dump(self, value: str) -> str:
        return str(value)


class IntegerSerializer(Serializer[int]):
    regex = re.compile(r"-?[0-9]+")
    type = int

    def load(self, value: str) -> int:
        return int(value)

    def dump(self, value: int) -> str:
        return str(value)


class FloatSerializer(Serializer[float]):
    regex = re.compile(r"-?[0-9]+(.[0-9]+)?")
    type = float

    def load(self, value: str) -> float:
        return float(value)

    def dump(self, value: float) -> str:
        return f"{value:0.10f}".rstrip("0").rstrip(".")


class DecimalSerializer(Serializer[decimal.Decimal]):
    regex = re.compile(r"-?[0-9]+(.[0-9]+)?")
    type = decimal.Decimal

    def load(self, value: str) -> decimal.Decimal:
        return decimal.Decimal(value)

    def dump(self, value: decimal.Decimal) -> str:
        return str(value)


class UUIDSerializer(Serializer[uuid.UUID]):
    regex = re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
    type = uuid.UUID

    def load(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def dump(self, value: uuid.UUID) -> str:
        return str(value)


@dataclasses.dataclass
class _Fragment(abc.ABC):
    REGEX: t.ClassVar[re.Pattern] = re.compile(
        r"""
        (?P<parameter>{(?P<parameter_name>[a-zA-Z_][a-zA-Z0-9_]*)(?::(?P<parameter_type>[a-zA-Z_][a-zA-Z0-9_]*)?)?}) |
        (?P<constant>.*)
        """,
        re.X,
    )

    value: str
    type: FragmentType

    @classmethod
    def build(cls, fragment: str) -> "_Fragment":
        match = cls.REGEX.match(fragment)

        assert match

        if match.group("parameter"):
            return _FragmentParameter(
                value=fragment,
                type=t.cast(FragmentType, match.group("parameter_type") or "str"),
                name=match.group("parameter_name"),
            )
        else:
            return _FragmentConstant(value=fragment, type="constant")

    @property
    @abc.abstractmethod
    def regex(self) -> re.Pattern: ...

    @property
    @abc.abstractmethod
    def template(self) -> str: ...


@dataclasses.dataclass
class _FragmentConstant(_Fragment):
    @property
    def regex(self) -> re.Pattern:
        return re.compile(self.value)

    @property
    def template(self) -> str:
        return self.value


@dataclasses.dataclass
class _FragmentParameter(_Fragment):
    SERIALIZERS: t.ClassVar[dict[FragmentType, Serializer]] = {
        "str": StringSerializer(),
        "int": IntegerSerializer(),
        "float": FloatSerializer(),
        "decimal": DecimalSerializer(),
        "uuid": UUIDSerializer(),
    }

    name: str = dataclasses.field(repr=False, hash=False, compare=False)
    serializer: Serializer = dataclasses.field(init=False, repr=False, hash=False, compare=False)

    def __post_init__(self):
        try:
            self.serializer = self.SERIALIZERS[self.type]
        except KeyError:
            raise ValueError(f"Unknown path serializer '{self.type}'")

    @property
    def regex(self) -> re.Pattern:
        return re.compile(rf"(?P<{self.name}>{self.serializer.regex.pattern})")

    @property
    def template(self) -> str:
        return f"{{{self.name}}}"


class _Match(enum.Enum):
    exact = enum.auto()
    partial = enum.auto()
    none = enum.auto()


@dataclasses.dataclass
class _MatchResult:
    match: _Match
    parameters: t.Optional[dict[str, t.Any]]
    matched: t.Optional[str]
    unmatched: t.Optional[str]


@dataclasses.dataclass
class _BuildResult:
    path: str
    unused: dict[str, t.Any]


class Path:
    Match = _Match

    def __init__(self, path: t.Union[str, "Path"]):
        """URL path with a regex to allow path params as placeholders.

        Given a path string like: "/foo/{bar:str}"
        path:        "/foo/{bar:str}"
        template:    "/foo/{bar}"
        regex:       "^/foo/(?P<bar>[^/]+)$"

        :param path: URL path.
        """
        if isinstance(path, Path):
            self.path = path.path
            self._fragments = path._fragments
            self._parameters = path._parameters
            self._regex = path._regex
            self._template = path._template
        else:
            if path != "" and not path.startswith("/"):
                raise ValueError("Path must starts with '/'")

            self.path = path
            self._fragments = [_Fragment.build(x) for x in path.strip("/").split("/")]
            self._parameters: dict[str, _FragmentParameter] = {
                f.name: f for f in self._fragments if isinstance(f, _FragmentParameter)
            }

            starting_slash = "/" if path != "" else ""
            trailing_slash = "/" if self.path != "/" and self.path.endswith("/") else ""

            fragments_templates = "/".join(f.template for f in self._fragments)
            self._template = f"{starting_slash}{fragments_templates}{trailing_slash}"

            fragments_regex = "/".join(f.regex.pattern for f in self._fragments)
            self._regex = re.compile(
                rf"^(?P<__matched__>{starting_slash}{fragments_regex}{trailing_slash})(?P<__unmatched__>.*)$"
            )

    @property
    def parameters(self) -> dict[str, type]:
        return {f.name: f.serializer.type for f in self._parameters.values()}

    def match(self, path: t.Union[str, "Path"]) -> _MatchResult:
        """Check if given path matches with current object.

        :param path: Path to match
        :return: Matching result, parameters serialized values and matching parts of the path.
        """
        if (match := self._regex.match(str(path))) is None:
            return _MatchResult(self.Match.none, None, None, None)

        return _MatchResult(
            match=self.Match.partial if match.group("__unmatched__") else self.Match.exact,
            parameters={
                k: self._parameters[k].serializer.load(v)
                for k, v in match.groupdict().items()
                if k not in ("__matched__", "__unmatched__")
            },
            matched=match.group("__matched__") or None,
            unmatched=match.group("__unmatched__") or None,
        )

    def build(self, **params: t.Any) -> _BuildResult:
        """Build a path by completing param placeholders with given values.

        :param params: Param values.
        :return: Built path and unused params.
        """
        if not set(self._parameters.keys()) <= set(params.keys()):
            formatted_params = ", ".join(f"'{x}'" for x in self._parameters.keys())
            raise ValueError(f"Wrong params, expected: {formatted_params}.")

        return _BuildResult(
            path=self._template.format(
                **{k: self._parameters[k].serializer.dump(v) for k, v in params.items() if k in self._parameters}
            ),
            unused={k: v for k, v in params.items() if k not in self._parameters},
        )

    def __bool__(self) -> bool:
        return self.path != ""

    def __eq__(self, other, /) -> bool:
        return isinstance(other, Path) and self.path.__eq__(other.path) or self.path.__eq__(other)

    def __str__(self) -> str:
        return self._template.__str__()

    def __repr__(self) -> str:
        return self.path.__repr__()

    def __truediv__(self, other: t.Union["Path", str]) -> "Path":
        if isinstance(other, Path):
            a, b = self.path.rstrip("/"), other.path.lstrip("/")
        elif isinstance(other, str):
            a, b = self.path.rstrip("/"), other.lstrip("/")
        else:
            raise TypeError(f"Can only concatenate str or {self.__class__.__name__} to {self.__class__.__name__}")

        return Path(f"{a}/{b}")

    def __rtruediv__(self, other: t.Union["Path", str]) -> "Path":
        if isinstance(other, Path):
            a, b = other.path.rstrip("/"), self.path.lstrip("/")  # pragma: no cover # covered by __truediv__
        elif isinstance(other, str):
            a, b = other.rstrip("/"), self.path.lstrip("/")
        else:
            raise TypeError(f"Can only concatenate str or {self.__class__.__name__} to {self.__class__.__name__}")

        return Path(f"{a}/{b}")

    def __itruediv__(self, other: t.Union["Path", str]) -> "Path":
        path = self / other
        self.path = path.path
        self._fragments = path._fragments
        self._parameters = path._parameters
        self._regex = path._regex
        self._template = path._template

        return self


@dataclasses.dataclass
class URL:
    scheme: str
    netloc: str
    path: Path
    params: str
    query: str
    fragment: str

    def __init__(self, url: t.Union[str, "URL"] = "", /, **kwargs):
        """URL object.

        :param url: URL string to be parsed.
        :param kwargs: Individual components to replace those parsed from string.
        """
        parsed_url = urllib.parse.urlparse(url)._replace(**kwargs) if isinstance(url, str) else url
        self.scheme = parsed_url.scheme
        self.netloc = parsed_url.netloc
        self.path = Path(parsed_url.path)
        self.params = parsed_url.params
        self.query = parsed_url.query
        self.fragment = parsed_url.fragment

    @property
    def components(self) -> dict[str, t.Optional[str]]:
        """URL components map.

        :return: Components.
        """
        return {
            "scheme": self.scheme,
            "netloc": self.netloc,
            "path": str(self.path),
            "params": self.params,
            "query": self.query,
            "fragment": self.fragment,
        }

    @property
    def url(self) -> str:
        """Build URL string.

        :return: URL string.
        """
        return str(urllib.parse.urlunparse(tuple(self.components.values())))

    def __eq__(self, value, /) -> bool:
        return isinstance(value, URL) and self.url == value.url or self.url == value

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"URL('{self.url}')"



import abc

__all__ = ["AbstractRepository", "BaseRepository"]


class AbstractRepository(abc.ABC):
    """Abstract class for repositories."""

    def __init__(self, *args, **kwargs): ...


class BaseRepository(AbstractRepository):
    """Base class for repositories."""

    ...



import typing as t
from pathlib import Path
from types import ModuleType

from flama import exceptions, http, pagination, schemas, types
from flama.modules import Module
from flama.schemas.generator import SchemaGenerator

__all__ = ["SchemaModule"]

TEMPLATES_PATH = Path(__file__).parents[1] / "templates"


class SchemaModule(Module):
    name = "schema"

    def __init__(self, openapi: types.OpenAPISpec, *, schema: t.Optional[str] = None, docs: t.Optional[str] = None):
        super().__init__()

        if docs and not schema:
            raise exceptions.ApplicationError("Docs endpoint needs schema endpoint to be active")

        # Schema definitions
        self.schemas: dict[str, t.Any] = {}

        # Schema
        self.openapi = openapi
        self.schema_path = schema
        self.docs_path = docs

    def register_schema(self, name: str, schema: t.Any) -> None:
        """Register a new schema.

        :param name: Schema name.
        :param schema: Schema.
        """
        self.schemas[name] = schema

    @property
    def schema_generator(self) -> SchemaGenerator:
        """Build an API Schema Generator.

        :return: API Schema Generator.
        """
        self.schemas.update({**schemas.schemas.SCHEMAS, **pagination.paginator.schemas})
        return SchemaGenerator(spec=self.openapi, schemas=self.schemas)

    @property
    def schema(self) -> dict[str, t.Any]:
        """Generate the API schema.

        :return: API schema.
        """
        return self.schema_generator.get_api_schema(self.app.routes)

    @property
    def schema_library(self) -> ModuleType:
        """Global schema library.

        :return: Schema library module.
        """
        return schemas._module.lib

    @schema_library.setter
    def schema_library(self, library: str) -> None:
        """Globally set the schema library.

        :param library: Schema library to be used.
        """
        schemas._module.setup(library)

    def add_routes(self) -> None:
        """Add schema and docs routes."""
        if self.schema_path:
            self.app.add_route(self.schema_path, self.schema_view, methods=["GET"], include_in_schema=False)
        if self.docs_path:
            self.app.add_route(self.docs_path, self.docs_view, methods=["GET"], include_in_schema=False)

    def schema_view(self) -> http.OpenAPIResponse:
        return http.OpenAPIResponse(self.schema)

    def docs_view(self) -> http.HTMLResponse:
        return http._FlamaTemplateResponse("schemas/docs.html", {"url": self.schema_path})
